"""SDMX-CSV 2.x formats."""

from dataclasses import dataclass, field
from enum import Enum, auto

from sdmx.format.csv.common import CSVFormat, CSVFormatOptions


class Keys(Enum):
    """SDMX-CSV 2.x 'keys' parameter."""

    #: No related columns.
    none = auto()

    #: Both :attr:`obs` and :attr:`series`.
    both = auto()

    #: Include ``OBS_KEY`` column with key values for all dimension(s).
    obs = auto()

    #: Include ``SERIES_KEY`` column with key values for all dimension(s) *except* the
    #: one(s) attached to each observation.
    series = auto()


class FORMAT(CSVFormat):
    version = "2.0.0"


@dataclass
class FormatOptions(CSVFormatOptions):
    """SDMX-CSV 2.x format options."""

    format = FORMAT

    #: SDMX-CSV 'keys' parameter.
    keys: Keys = Keys.none

    #: “Custom columns” detected by :meth:`.Reader.inspect_header`.
    custom_columns: list[bytes] = field(default_factory=list)

    #: CSV field delimiter.
    delimiter: str = ","

    #: SDMX-CSV “sub-field” delimiter.
    delimiter_sub: str = ""



import re
from typing import TYPE_CHECKING, Any, Optional, Union

import pytest

from sdmx import Resource
from sdmx.rest.common import PathParameter, PositiveIntParam, QueryParameter

if TYPE_CHECKING:
    import _pytest


class TestResource:
    @pytest.mark.parametrize(
        "value, expected",
        [
            # Some values are mapped to different class names
            (Resource.dataflow, "DataflowDefinition"),
            (Resource.datastructure, "DataStructureDefinition"),
            # Other values pass through
            (Resource.agencyscheme, Resource.agencyscheme.value),
        ],
    )
    def test_class_name(self, value, expected):
        assert expected == Resource.class_name(value)

    def test_describe(self):
        assert re.fullmatch(
            "{actualconstraint .* vtlmappingscheme}", Resource.describe()
        )


class TestPathParameter:
    @pytest.fixture
    def p(self):
        yield PathParameter("name", {"foo", "bar"})

    def test_handle(self, p):
        p.handle({"name": "foo"})

        with pytest.raises(ValueError, match="Missing required parameter 'name'"):
            p.handle({})

        with pytest.raises(ValueError, match="name='baz' not among "):
            p.handle({"name": "baz"})


class TestQueryParameter:
    @pytest.fixture
    def p(self):
        yield QueryParameter("snake_name", {1, 2})

    def test_handle(self, p):
        params = {"snake_name": None}
        p.handle(params)
        assert 0 == len(params)

        with pytest.raises(ValueError, match="Cannot give both"):
            p.handle({"snake_name": 1, "snakeName": 2})

        with pytest.raises(ValueError, match="snake_name=3 not among "):
            p.handle({"snakeName": 3})


class TestPositiveIntParam:
    @pytest.fixture
    def p(self):
        yield PositiveIntParam("name")

    def test_handle(self, p):
        with pytest.raises(ValueError, match="must be positive integer; got"):
            p.handle({"name": -3})

        p.handle({"name": 1})


_S = "?startPeriod=2024-02-12"

R = Resource
PARAMS: tuple[
    Union[
        tuple[Resource, dict[str, Any], Optional[str], Optional[str]],
        "_pytest.mark.ParameterSet",
    ],
    ...,
] = (
    # (R.actualconstraint, {}, "actualconstraint/ID0", "actualconstraint/ID0"),
    (
        R.agencyscheme,
        {},
        "agencyscheme/A0/ID0/latest",
        "structure/agencyscheme/A0/ID0/+",
    ),
    (
        R.allowedconstraint,
        {},
        "allowedconstraint/A0/ID0/latest",
        "structure/allowedconstraint/A0/ID0/+",
    ),
    (
        R.attachementconstraint,
        {},
        "attachementconstraint/A0/ID0/latest",
        "structure/attachementconstraint/A0/ID0/+",
    ),
    # v21 only
    (
        R.availableconstraint,
        dict(key="111+134.PCPIHA_IX.*"),
        "availableconstraint/ID0/111+134.PCPIHA_IX.*",
        None,
    ),
    # v30 only
    (
        R.availableconstraint,
        dict(key="111+134.PCPIHA_IX.*", component_id="*", context="dataflow"),
        None,
        "availability/dataflow/A0/ID0/+/111+134.PCPIHA_IX.*/*",
    ),
    (
        R.categorisation,
        {},
        "categorisation/A0/ID0/latest",
        "structure/categorisation/A0/ID0/+",
    ),
    (
        R.categoryscheme,
        {},
        "categoryscheme/A0/ID0/latest?references=parentsandsiblings",
        "structure/categoryscheme/A0/ID0/+",
    ),
    # All other parameters explicit
    (R.codelist, {}, "codelist/A0/ID0/latest", "structure/codelist/A0/ID0/+"),
    # Providing explicit version
    (
        R.codelist,
        {"version": "1.0"},
        "codelist/A0/ID0/1.0",
        "structure/codelist/A0/ID0/1.0",
    ),
    (
        R.codelist,
        {"version": "1.0"},
        "codelist/A0/ID0/1.0",
        "structure/codelist/A0/ID0/1.0",
    ),
    (
        R.conceptscheme,
        {},
        "conceptscheme/A0/ID0/latest",
        "structure/conceptscheme/A0/ID0/+",
    ),
    (
        R.contentconstraint,
        {},
        "contentconstraint/A0/ID0/latest",
        "structure/contentconstraint/A0/ID0/+",
    ),
    (
        R.customtypescheme,
        {},
        "customtypescheme/A0/ID0/latest",
        "structure/customtypescheme/A0/ID0/+",
    ),
    # No extra parameters
    (R.data, {}, "data/ID0", "data/*/A0/ID0/+"),
    # Different ways of expressing the same query
    (R.data, dict(start_period="2024-02-12"), f"data/ID0{_S}", None),
    (R.data, dict(startPeriod="2024-02-12"), f"data/ID0{_S}", None),
    (R.data, dict(params=dict(start_period="2024-02-12")), f"data/ID0{_S}", None),
    (R.data, dict(params=dict(startPeriod="2024-02-12")), f"data/ID0{_S}", None),
    (
        R.dataconsumerscheme,
        {},
        "dataconsumerscheme/A0/ID0/latest",
        "structure/dataconsumerscheme/A0/ID0/+",
    ),
    (
        R.dataflow,
        {},
        "dataflow/A0/ID0/latest?references=all",
        "structure/dataflow/A0/ID0/+",
    ),
    (
        R.dataproviderscheme,
        {},
        "dataproviderscheme/A0/ID0/latest",
        "structure/dataproviderscheme/A0/ID0/+",
    ),
    (
        R.datastructure,
        {},
        "datastructure/A0/ID0/latest?references=all",
        "structure/datastructure/A0/ID0/+",
    ),
    (
        R.hierarchicalcodelist,
        {},
        "hierarchicalcodelist/A0/ID0/latest",
        "structure/hierarchicalcodelist/A0/ID0/+",
    ),
    (
        R.metadata,
        dict(
            provider_id="*",
        ),
        # "metadata/ID0",
        None,  # Not implemented; no reference implementation available
        "metadata/metadataset/*/ID0/+",
    ),
    (
        R.metadataflow,
        {},
        "metadataflow/A0/ID0/latest",
        "structure/metadataflow/A0/ID0/+",
    ),
    (
        R.metadatastructure,
        {},
        "metadatastructure/A0/ID0/latest",
        "structure/metadatastructure/A0/ID0/+",
    ),
    (
        R.namepersonalisationscheme,
        {},
        "namepersonalisationscheme/A0/ID0/latest",
        "structure/namepersonalisationscheme/A0/ID0/+",
    ),
    (
        R.organisationscheme,
        {},
        "organisationscheme/A0/ID0/latest",
        "structure/organisationscheme/A0/ID0/+",
    ),
    (
        R.organisationunitscheme,
        {},
        "organisationunitscheme/A0/ID0/latest",
        "structure/organisationunitscheme/A0/ID0/+",
    ),
    (R.process, {}, "process/A0/ID0/latest", "structure/process/A0/ID0/+"),
    (
        R.provisionagreement,
        {},
        "provisionagreement/A0/ID0/latest",
        "structure/provisionagreement/A0/ID0/+",
    ),
    (
        R.registration,
        dict(context="dataflow", agency_id="A0"),
        None,
        "registration/dataflow/A0/ID0/+",
    ),
    pytest.param(
        R.registration,
        dict(agency_id="A0", provider_id="P0"),
        None,
        "registration/provider/A0/P0",
        marks=pytest.mark.xfail(raises=ValueError),
    ),
    (R.registration, {}, None, "registration/id/ID0"),
    (
        R.reportingtaxonomy,
        {},
        "reportingtaxonomy/A0/ID0/latest",
        "structure/reportingtaxonomy/A0/ID0/+",
    ),
    (
        R.rulesetscheme,
        {},
        "rulesetscheme/A0/ID0/latest",
        "structure/rulesetscheme/A0/ID0/+",
    ),
    (
        R.schema,
        dict(context="provisionagreement"),
        "schema/provisionagreement/A0/ID0/latest",
        "schema/provisionagreement/A0/ID0/+",
    ),
    # (R.structure, {}, "structure/A0/ID0/latest", None),
    (
        R.structureset,
        {},
        "structureset/A0/ID0/latest",
        "structure/structureset/A0/ID0/+",
    ),
    (
        R.transformationscheme,
        {},
        "transformationscheme/A0/ID0/latest",
        "structure/transformationscheme/A0/ID0/+",
    ),
    (
        R.userdefinedoperatorscheme,
        {},
        "userdefinedoperatorscheme/A0/ID0/latest",
        "structure/userdefinedoperatorscheme/A0/ID0/+",
    ),
    (
        R.vtlmappingscheme,
        {},
        "vtlmappingscheme/A0/ID0/latest",
        "structure/vtlmappingscheme/A0/ID0/+",
    ),
)

PARAMS_INVALID = (
    # Providing an invalid parameter
    (R.codelist, {"foo": "bar"}, ValueError, "Unexpected/unhandled parameters"),
    # Providing duplicate path and query parameters
    (
        R.codelist,
        {"foo": "bar", "params": {"foo": "baz"}},
        ValueError,
        "Duplicate values for",
    ),
    # Positive integer parameter must be positive integer
    (
        R.data,
        {"first_n_observations": -1},
        ValueError,
        "must be positive integer; got -1",
    ),
)


class URLTests:
    """Common fixtures for testing URL classes."""

    expected_index: int

    @pytest.fixture
    def source(self):
        from sdmx.source import Source

        return Source(id="A0", url="https://example.com", name="Test source")

    @pytest.mark.parametrize("resource_type, kw, expected0, expected1", PARAMS)
    def test_join(self, URL, source, resource_type, kw, expected0, expected1) -> None:
        expected = [expected0, expected1][self.expected_index]
        if expected is None:
            return

        # Instance can be created
        u = URL(source, resource_type, resource_id="ID0", **kw)

        # Constructed URL is as expected
        assert f"https://example.com/{expected}" == u.join()

    @pytest.mark.parametrize("resource_type, kw, exc_type, exc_re", PARAMS_INVALID)
    def test_join_invalid(self, URL, source, resource_type, kw, exc_type, exc_re):
        with pytest.raises(exc_type, match=exc_re):
            URL(source, resource_type, resource_id="ID0", **kw)


class TestURLv21(URLTests):
    """Construction of SDMX REST API 2.1 URLs."""

    expected_index = 0

    @pytest.fixture
    def URL(self):
        import sdmx.rest.v21

        return sdmx.rest.v21.URL


class TestURLv30(URLTests):
    """Construction of SDMX REST API 3.0.0 URLs."""

    expected_index = 1

    @pytest.fixture
    def URL(self):
        import sdmx.rest.v30

        return sdmx.rest.v30.URL



"""Utilities for working with :mod:`requests` and related packages."""

from importlib.util import find_spec
from typing import TYPE_CHECKING, Optional, TypedDict, Union

import urllib3
from requests import PreparedRequest, Response
from requests.adapters import BaseAdapter

#: :any:`True` if :class:`requests_cache` is installed.
HAS_REQUESTS_CACHE = bool(find_spec("requests_cache"))

if HAS_REQUESTS_CACHE:
    from requests_cache import CacheMixin
else:  # pragma: no cover

    class CacheMixin:  # type: ignore [no-redef]
        """Null parent class for sdmx.session.Session."""


if TYPE_CHECKING:
    import http.cookiejar

    from requests import Session

__all__ = [
    "HAS_REQUESTS_CACHE",
    "CacheMixin",
    "OfflineAdapter",
    "SessionAttrs",
    "offline",
    "save_response",
]


class OfflineAdapter(BaseAdapter):
    """A request Adapter that raises :class:`RuntimeError` for every request.

    See also
    --------
    offline
    """

    def send(self, request, **kwargs):
        raise RuntimeError(f"Attempted query in offline mode for URL:\n{request.url}")


class SessionAttrs(TypedDict):
    """Attributes of :class:`.requests.Session`.

    These are not available from :mod:`requests` itself, thus recorded here for use in
    :meth:`sdmx.session.Session.__init__`.
    """

    adapters: dict
    auth: Optional[object]
    cert: Union[str, tuple[str, str]]
    cookies: "http.cookiejar.CookieJar"
    headers: dict
    hooks: dict
    max_redirects: int
    params: dict
    proxies: dict
    stream: bool
    trust_env: bool
    verify: bool


def offline(s) -> None:
    """Make session `s` behave as if offline.

    Replace all of the :attr:`Session.adapters` of `s` with instances of
    :class:`OfflineAdapter`. This has the effect that any request made through `s` that
    is not handled in some other way (for instance, by :mod:`requests_cache`) will raise
    :class:`RuntimeError`.
    """
    a = OfflineAdapter()
    for k in s.adapters.keys():
        s.adapters[k] = a


def save_response(
    session: "Session", method: str, url: str, content: bytes, headers: dict
) -> None:
    """Store a response in the cache of `session`.

    If :mod:`requests_cache` is not available, this has no effect.
    """
    if not hasattr(session, "cache"):  # pragma: no cover
        return

    # Response object and its direct attributes
    resp = Response()
    resp._content = content
    resp.status_code = 200
    resp.headers.update(headers)
    resp.url = url

    # Subsidiary objects: 'raw' (urllib3) response
    resp.raw = urllib3.HTTPResponse(request_url=url)

    # PreparedRequest
    req = resp.request = PreparedRequest()
    req.prepare(method=method, url=url)

    session.cache.save_response(resp)



"""SDMX-JSON v2.1 reader"""

import json
import logging
from typing import Mapping, MutableMapping
from warnings import warn

from dateutil.parser import isoparse

from sdmx.format import Version, list_media_types
from sdmx.message import DataMessage, Header
from sdmx.model import common
from sdmx.model import v21 as model
from sdmx.model.common import Concept
from sdmx.model.internationalstring import InternationalString
from sdmx.model.v21 import (
    ActionType,
    AllDimensions,
    AttributeValue,
    DataSet,
    Key,
    KeyValue,
    Observation,
    SeriesKey,
)
from sdmx.reader.base import BaseReader

log = logging.getLogger(__name__)


class Reader(BaseReader):
    """Read SDMX-JSON and expose it as instances from :mod:`sdmx.model`."""

    binary_content_startswith = b"{"
    media_types = list_media_types(base="json", version=Version["1.0.0"])
    suffixes = [".json"]

    @classmethod
    def detect(cls, content):
        warn(
            "Reader.detect(bytes); use Converter.handles() instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return content.startswith(cls.binary_content_startswith)

    def convert(self, data, structure=None, **kwargs):  # noqa: C901  TODO reduce complexity 15 → ≤10
        # Initialize message instance
        msg = DataMessage()

        dsd = self._handle_deprecated_kwarg(structure, kwargs)
        if dsd:  # pragma: no cover
            # Store explicit DSD, if any
            msg.dataflow.structure = dsd

        # Read JSON
        data.default_size = -1
        tree = json.load(data)

        # Read the header
        # TODO handle KeyError here
        try:
            # SDMX-JSON 1.0 (SDMX 2.1)
            elem = tree["header"]
        except KeyError:
            # SDMX-JSON 2.0 (SDMX 3.0.0)
            elem = tree["meta"]

        msg.header = Header(
            id=elem["id"],
            prepared=isoparse(elem["prepared"]),
            sender=_org(elem["sender"], cls=common.Agency),
        )

        # pre-fetch some structures for efficient use in series and obs
        try:
            # SDMX-JSON 1.0 (SDMX 2.1)
            structure = tree["structure"]
        except KeyError:
            # SDMX-JSON 2.0 (SDMX 3.0.0)
            structure = tree["data"]["structures"]

        # Read dimensions and values
        self._dim_level = dict()
        self._dim_values = dict()
        for level_name, level in structure["dimensions"].items():
            for elem in level:
                # Create the Dimension
                d = msg.structure.dimensions.getdefault(
                    id=elem["id"], order=elem.get("keyPosition", -1)
                )

                # Record the level it appears at
                self._dim_level[d] = level_name

                # Record values
                self._dim_values[d] = list()
                for value in elem.get("values", []):
                    self._dim_values[d].append(KeyValue(id=d.id, value=value["id"]))

        # Assign an order to an implicit dimension
        for d in msg.structure.dimensions:
            if d.order == -1:
                d.order = len(msg.structure.dimensions)

        # Determine the dimension at the observation level
        if all([level == "observation" for level in self._dim_level.values()]):
            dim_at_obs = AllDimensions
        else:
            dim_at_obs = [
                dim for dim, level in self._dim_level.items() if level == "observation"
            ]

        msg.observation_dimension = dim_at_obs

        # Read attributes and values
        self._attr_level = dict()
        self._attr_values = dict()
        for level_name, level in structure["attributes"].items():
            for attr in level:
                # Create a DataAttribute in the DSD
                da = msg.structure.attributes.getdefault(
                    id=attr["id"], concept_identity=Concept(name=attr["name"])
                )

                # Record its values
                values = []
                for v in attr.get("values", []):
                    values.append(
                        AttributeValue(
                            value=model.Code(**v) if "id" in v else v["name"],
                            value_for=da,
                        )
                    )

                # Handle https://github.com/khaeru/sdmx/issues/64: a DataAttribute with
                # no values cannot be referenced, and is assumed to be erroneously
                # included.
                if not len(values):
                    log.warning(f"No AttributeValues for attribute {repr(da)}; discard")

                    # Remove the DataAttribute
                    idx = msg.structure.attributes.components.index(da)
                    msg.structure.attributes.components.pop(idx)

                    continue

                self._attr_values[da] = values

                # Record the level it appears at
                self._attr_level[da] = level_name

        self.msg = msg

        # Make a SeriesKey for Observations in this DataSet
        ds_key = self._make_key("dataSet")

        # Read DataSets
        for ds in tree["dataSets"]:
            msg.data.append(self.read_dataset(ds, ds_key))

        return msg

    def read_dataset(self, root, ds_key):
        ds = DataSet(
            action=ActionType[root["action"].lower()],
            valid_from=root.get("validFrom", None),
        )

        # Process series
        for key_values, elem in root.get("series", {}).items():
            series_key = self._make_key("series", key_values, base=ds_key)
            series_key.attrib = self._make_attrs("series", root.get("attributes", []))
            ds.add_obs(self.read_obs(elem, series_key=series_key), series_key)

        # Process bare observations
        ds.add_obs(self.read_obs(root, base_key=ds_key))

        return ds

    def read_obs(self, root, series_key=None, base_key=None):
        for key, elem in root.get("observations", {}).items():
            value = elem.pop(0) if len(elem) else None
            o = Observation(
                series_key=series_key,
                dimension=self._make_key("observation", key, base=base_key),
                value=value,
                attached_attribute=self._make_attrs("observation", elem),
            )
            yield o

    def _make_key(self, level, value=None, base=None):
        """Convert a string observation key *value* to a Key or subclass.

        SDMXJSON observations have keys like '2' or '3:4', consisting of colon
        (':') separated indices. Each index refers to one of the values given
        in the DSD for an observation-level dimension.

        KeyValues from any *base* Key are copied, and the new values appended.
        *level* species whether a 'series' or 'observation' Key is returned.
        """
        # Instance of the proper class
        key = {"dataSet": Key, "series": SeriesKey, "observation": Key}[level]()

        if base:
            key.values.update(base.values)

        # Dimensions at the appropriate level
        dims = [d for d in self.msg.structure.dimensions if self._dim_level[d] == level]

        # Dimensions specified at the dataSet level have only one value, so
        # pre-fill this
        value = ":".join(["0"] * len(dims)) if value is None else value

        if len(value):
            # Iterate over key indices and the corresponding dimensions
            for index, dim in zip(map(int, value.split(":")), dims):
                # Look up the value and assign to the Key
                key[dim.id] = self._dim_values[dim][index]

        # Order the key
        return self.msg.structure.dimensions.order_key(key)

    def _make_attrs(self, level, values):
        """Convert integer attribute indices to an iterable of AttributeValues.

        'level' must be one of 'dataSet', 'series', or 'observation'.
        """
        attrs = [
            a for a in self.msg.structure.attributes if self._attr_level[a] == level
        ]
        result = {}
        for index, attr in zip(values, attrs):
            if index is None:
                continue
            av = self._attr_values[attr][index]
            result[av.value_for.id] = av
        return result


def _org(elem: MutableMapping, cls=common.Organisation) -> common.Organisation:
    try:
        elem["contact"] = list(map(_contact, elem.pop("contacts")))
    except KeyError:
        pass
    return cls(**maybe_parse_is(elem, "name"))


def _contact(elem: Mapping) -> common.Contact:
    data = depluralize(
        maybe_parse_is(elem, "department name role"),
        "email fax|faxe telephone uri x400",
    )
    if "id" in data:
        log.debug(f"Discard unsupported Contact(id={data.pop('id')!r})")
    try:
        data["org_unit"] = data.pop("department")
    except KeyError:  # pragma: no cover
        pass
    try:
        data["responsibility"] = data.pop("role")
    except KeyError:  # pragma: no cover
        pass
    return common.Contact(**data)


def depluralize(elem: Mapping, names: str) -> MutableMapping:
    """Map plural `names` in `elem` to singular.

    `names` is a whitespace-separated sequence of strings; each string is either the
    singular form and base for the plural form ("cat" → "cat" and "cats") or a pipe (|)
    separating the singular and plural base ("fox|foxe" → "fox", "foxes").
    """
    result = dict(elem)
    for name in names.split():
        info = name.split("|")
        try:
            result[info[0]] = result.pop(f"{info[-1]}s")
        except KeyError:  # pragma: no cover
            pass
    return result


def maybe_parse_is(elem: Mapping, names: str) -> Mapping:
    """Parse international string(s) with given `names` appearing on `elem`, if any."""
    result = dict(elem)
    for name in names.split():
        if name in result and f"{name}s" in result:
            result[name] = InternationalString(result.pop(f"{name}s"))
    return result



import pytest

from sdmx import message
from sdmx.model.common import Agency, Codelist
from sdmx.model.v21 import Annotation

CL_ITEMS = [
    dict(id="A", name={"en": "Average of observations through period"}),
    dict(id="B", name={"en": "Beginning of period"}),
    dict(id="B1", name={"en": "Child code of B"}, parent="B"),
]


@pytest.fixture
def codelist():
    """A Codelist for writer testing."""
    ECB = Agency(id="ECB")

    cl = Codelist(
        id="CL_COLLECTION",
        version="1.0",
        is_final=False,
        is_external_reference=False,
        maintainer=ECB,
        name={"en": "Collection indicator code list"},
    )

    # Add items
    for info in CL_ITEMS:
        # For B1, this also adds a hierarchical relationship to B
        code = cl.setdefault(**info)
        # FIXME setdefault should do this automatically
        if "parent" not in info:
            code.parent = cl

    # Add an annotation
    cl.items["A"].annotations.append(
        Annotation(id="A1", type="NOTE", text={"en": "Text annotation on Code A."})
    )

    return cl


# Classes from sdmx.message


@pytest.fixture
def footer():
    """A :class:`.Footer` for writer testing."""
    return message.Footer(
        code=123,
        severity="information",
        text=["Here is one string.", "Here is another."],
    )


@pytest.fixture
def structuremessage(codelist, footer):
    """A :class:`.StructureMessage.` for writer testing."""
    return message.StructureMessage(codelist={codelist.id: codelist}, footer=footer)


@pytest.fixture
def datamessage(footer):
    return message.DataMessage(footer=footer)


@pytest.fixture
def errormessage(footer):
    return message.ErrorMessage(footer=footer)



from collections.abc import Callable
from typing import Any


class Converter:
    """Base class for conversion to or from :mod:`sdmx` objects."""

    @classmethod
    def handles(cls, data: Any, kwargs: dict) -> bool:
        """Return :any:`True` if the class can convert `data` using `kwargs`."""
        return False

    def convert(self, data: Any, **kwargs) -> Any:
        """Convert `data`."""
        raise NotImplementedError


class DispatchConverter(Converter):
    """Base class for recursive converters.

    Usage:

    - Create a subclass of this class.
    - Use :meth:`register` (in the same manner as Python's built-in
      :func:`functools.singledispatch`) to decorate functions that convert certain types
      of :mod:`sdmx.model` or :mod:`sdmx.message` objects.
    - Call :meth:`convert` (including from inside other functions) to recursively
      convert objects.

    Example
    -------
    >>> from sdmx.convert.common import DispatchConverter
    >>> class CustomConverter(DispatchConverter):
    ...     pass

    >>> @CustomConverter.register
    ... def _(c: "CustomConverter", obj: sdmx.model.ItemScheme):
    ...     ... code to convert an ItemScheme ...
    ...     return result

    >>> @CustomConverter.register
    ... def _(c: "CustomConverter", obj: sdmx.model.Codelist):
    ...     ... code to convert a Codelist ...
    ...     return result
    """

    _registry: dict[type, Callable]

    def convert(self, obj, **kwargs):
        # Use either type(obj) or a parent type to retrieve a conversion function
        for i, cls in enumerate(type(obj).mro()):
            try:
                func = self._registry[cls]
            except KeyError:
                continue
            else:
                if i:  # Some superclass of type(obj) matched → cache for future calls
                    self._registry[type(obj)] = func
                break

        try:
            return func(self, obj, **kwargs)
        except UnboundLocalError:  # pragma: no cover
            raise NotImplementedError(
                f"Convert {type(obj)} using {type(self).__name__}"
            ) from None

    @classmethod
    def register(cls, func: "Callable"):
        """Register `func` as a conversion function.

        `func` must have an argument named `obj` that is annotated with a particular
        type.
        """
        try:
            registry = getattr(cls, "_registry")
        except AttributeError:
            # First call → registry does not exist → create it
            registry = dict()
            setattr(cls, "_registry", registry)

        # Register `func` for the class of the `obj` argument
        registry[getattr(func, "__annotations__")["obj"]] = func

        return func



import re
from datetime import datetime, timedelta, timezone
from io import BytesIO
from itertools import chain
from typing import cast

import pandas
import pytest
from lxml import etree

import sdmx
import sdmx.message
from sdmx import urn
from sdmx.format.xml import validate_xml
from sdmx.format.xml.v21 import qname
from sdmx.model import common, v21
from sdmx.model.v21 import ContentConstraint, Facet, FacetType, FacetValueType
from sdmx.reader.xml.v21 import Reader, XMLParseError
from sdmx.writer.xml import Element as E


class TestCategorisation:
    def test_0(self, specimen) -> None:
        """Check that :attr:`.Categorisation.target` is read as :class:`.Category`."""
        with specimen("ESTAT/esms-structure.xml") as f:
            msg = cast(sdmx.message.StructureMessage, sdmx.read_sdmx(f))

        c = msg.categorisation["DEMO_TOT"]

        assert isinstance(c.category, common.Category)  # Fails with v2.20.0


def test_read_xml_structure_insee(specimen):
    with specimen("IPI-2010-A21-structure.xml") as f:
        msg = sdmx.read_sdmx(f)

    # Same objects referenced
    assert id(msg.dataflow["IPI-2010-A21"].structure) == id(
        msg.structure["IPI-2010-A21"]
    )

    # Number of dimensions loaded correctly
    dsd = msg.structure["IPI-2010-A21"]
    assert len(dsd.dimensions) == 4


# Read structure-specific messages
def test_read_ss_xml(specimen):
    with specimen("M.USD.EUR.SP00.A.xml", opened=False) as f:
        msg_path = f
        dsd_path = f.parent / "structure.xml"

    # Read the DSD
    dsd = sdmx.read_sdmx(dsd_path).structure["ECB_EXR1"]

    # Read a data message
    msg = sdmx.read_sdmx(msg_path, structure=dsd)
    ds = msg.data[0]

    # The dataset in the message is structured by the DSD
    assert ds.structured_by is dsd

    # Structures referenced in the dataset are from the dsd
    s0_key = list(ds.series.keys())[0]

    # AttributeValue.value_for
    assert s0_key.attrib["DECIMALS"].value_for is dsd.attributes.get("DECIMALS")

    # SeriesKey.described_by
    assert s0_key.described_by is dsd.dimensions

    # Key.described_by
    assert ds.obs[0].key.described_by is dsd.dimensions

    # KeyValue.value_for
    assert ds.obs[0].key.values[0].value_for is dsd.dimensions.get("FREQ")

    # DSD information that is not in the data message can be looked up through
    # navigating object relationships
    TIME_FORMAT = s0_key.attrib["TIME_FORMAT"].value_for
    assert len(TIME_FORMAT.related_to.dimensions) == 5


def test_gh_078(specimen):
    """Test of https://github.com/khaeru/sdmx/issues/78.

    This required adding support for :xml:`<mes:Department>` and :xml:`<mes:Role>` to
    :mod:`.reader.xml`.
    """
    # Message can be read
    with specimen("WB/gh-78.xml") as f:
        msg = sdmx.read_sdmx(f)

    # Sender attributes are present and have the expected values
    for attr, text in (
        ("org_unit", "DECDG"),
        ("responsibility", "Support"),
    ):
        assert text == getattr(msg.header.sender.contact[0], attr).localizations["en"]


def test_gh_104(caplog, specimen):
    """Test of https://github.com/khaeru/sdmx/issues/104.

    See also
    --------
    .test_sources.TestISTAT.test_gh_104
    """
    # Read a DSD
    with specimen("22_289-structure.xml", opened=False) as f:
        dsd_path = f
        msg_path = f.parent / "22_289.xml"

    # Read the DSD, change its ID
    dsd = sdmx.read_sdmx(dsd_path).structure["DCIS_POPRES1"]
    dsd.id = "FOO"

    # Read a data message; use is logged
    sdmx.read_sdmx(msg_path, structure=dsd)
    assert re.match(
        r"Use provided <DataStructureDefinition IT1:FOO\(1\.0\): .* for "
        'structureRef="IT1_DCIS_POPRES1_1_0" not defined in message',
        caplog.messages[-1],
    )


def test_gh_116(specimen):
    """Test of https://github.com/khaeru/sdmx/issues/116.

    See also
    --------
    .test_sources.TestESTAT.test_gh_116
    """
    with specimen("ESTAT/GOV_10Q_GGNFA.xml") as f:
        msg = sdmx.read_sdmx(f)

    # Both versions of the GEO codelist are accessible in the message
    cl1 = msg.codelist["ESTAT:GEO(13.0)"]
    cl2 = msg.codelist["ESTAT:GEO(13.1)"]

    # cl1 is complete and items are available
    assert not cl1.is_partial and 0 < len(cl1)
    # cl2 is partial, and fewer codes are included than in cl1
    assert cl2.is_partial and 0 < len(cl2) < len(cl1)

    cl3 = msg.codelist["ESTAT:UNIT(15.1)"]
    cl4 = msg.codelist["ESTAT:UNIT(15.2)"]

    # cl3 is complete and items are available
    assert not cl3.is_partial and 0 < len(cl3)
    # cl4 is partial, and fewer codes are included than in cl1
    assert cl4.is_partial and 0 < len(cl4) < len(cl3)


def test_gh_142(specimen):
    """Test of https://github.com/khaeru/sdmx/issues/142."""
    with specimen("TEST/gh-142.xml") as f:
        msg = sdmx.read_sdmx(f)

    # Annotations, valid_from and valid_to properties stored on the Codelist *per se*
    cl = msg.codelist["CL_NAICS"]
    assert 3 == len(cl.annotations)
    assert "2021-01-24T08:00:00" == cl.valid_from
    assert "2021-09-24T08:00:00" == cl.valid_to

    # No annotations attached to any Code
    assert all(0 == len(code.annotations) for code in cl)


def test_gh_159():
    """Test of https://github.com/khaeru/sdmx/pull/159."""
    # Agency and contained Contact with distinct names
    elem = E(
        qname("str:Agency"),
        E(qname("com:Name"), "Foo Agency"),
        E(qname("str:Contact"), E(qname("com:Name"), "Jane Smith")),
        id="FOO",
    )

    # - Create a reader
    # - Convert to a file-like object compatible with convert()
    # - Parse the element
    # - Retrieve the resulting object
    reader = Reader()
    reader.convert(BytesIO(etree.tostring(elem)))
    obj = reader.pop_single(common.Agency)

    assert "Foo Agency" == str(obj.name)
    assert "Jane Smith" == str(obj.contact[0].name)


def test_gh_164(specimen):
    """Test of https://github.com/khaeru/sdmx/issues/164."""
    with specimen("IMF_STA/DSD_GFS.xml") as f:
        msg = sdmx.read_sdmx(f)

    # Retrieve objects directly from the message
    dsd = msg.structure["DSD_GFS"]
    IMF_CL_COUNTRY = msg.codelist["CL_COUNTRY"]
    IMF_CS_MASTER = msg.concept_scheme["IMF:CS_MASTER"]
    IMF_STA_CS_MASTER = msg.concept_scheme["IMF_STA:CS_MASTER"]

    # Component's concept identity is resolved within the correct concept scheme
    c = dsd.dimensions.get("COUNTRY").concept_identity
    cs = c.parent
    assert cs.urn.endswith("ConceptScheme=IMF:CS_MASTER(4.0.0)")
    assert urn.make(c).endswith("Concept=IMF:CS_MASTER(4.0.0).COUNTRY")

    # The reference is to the full concept scheme in the same message
    assert cs is IMF_CS_MASTER
    assert 56 == len(cs) == len(IMF_CS_MASTER)

    # Concept's core representation, per that scheme, can be retrieved
    assert c.core_representation.enumerated is IMF_CL_COUNTRY

    # Another component's concept identity is resolved within a concept scheme with the
    # same ID, but different maintainer and version:
    c = dsd.attributes.get("STATUS").concept_identity
    cs = c.parent
    assert cs.urn.endswith("ConceptScheme=IMF_STA:CS_MASTER(1.0.1)")
    assert urn.make(c).endswith("Concept=IMF_STA:CS_MASTER(1.0.1).STATUS")

    # The reference is to the full concept scheme in the same message
    assert cs is IMF_STA_CS_MASTER
    assert 12 == len(cs) == len(IMF_STA_CS_MASTER)

    # Concept's core representation as a text / string is parsed and stored
    facet = c.core_representation.non_enumerated[0]
    assert common.FacetValueType["string"] == facet.value_type

    # NoSpecifiedRelationship is correctly parsed
    da = dsd.attributes.get("FULL_CITATION")
    assert isinstance(da.related_to, v21.NoSpecifiedRelationship)


def test_gh_180(caplog, installed_schemas, specimen) -> None:
    """Test of https://github.com/khaeru/sdmx/issues/190."""
    with specimen("BIS/gh-180.xml") as f:
        # Message is not valid SDMX-ML
        assert False is validate_xml(f, installed_schemas)

        # Validation logs an error message regarding the non-standard class
        assert re.match(
            ".*attribute 'package'.*'publicationtable' is not an element of the set",
            caplog.messages[-2],
        )

        # Message can still be read
        f.seek(0)
        msg = sdmx.read_sdmx(f)
        assert isinstance(msg, sdmx.message.StructureMessage)

        # Reader logs a warning regarding the missing reference
        assert any(
            re.match("Cannot resolve reference to non-SDMX class", m)
            for m in caplog.messages
        )


def test_gh_199():
    """Test of https://github.com/khaeru/sdmx/issues/199."""
    import sdmx.format.xml.v21

    # Template for DSD URN
    URN = "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure=FOO:BAR({})"

    # Template for SDMX-ML data message
    CONTENT = """<?xml version="1.0" encoding="UTF-8"?>
<mes:StructureSpecificData
  xmlns:mes="http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message"
  xmlns:u="{}:ObsLevelDim:TIME_PERIOD">
  <u:DataSet>
    ...
  </u:DataSet>
</mes:StructureSpecificData>"""

    # Construct a URN and message; store as BytesIO
    urn1 = URN.format("1")
    dsd1 = v21.DataStructureDefinition(urn=urn1)
    f1 = BytesIO(CONTENT.format(urn1).encode())

    # Construct a *different* URN and message with this other URN mapped to the "u:" XML
    # namespace prefix
    urn2 = URN.format("2")
    dsd2 = v21.DataStructureDefinition(urn=urn2)
    f2 = BytesIO(CONTENT.format(urn2).encode())

    # First message can be parsed
    sdmx.read_sdmx(f1, structure=dsd1)

    # #199: raises XMLParseError/NotImplementedError
    sdmx.read_sdmx(f2, structure=dsd2)


def test_gh_205(caplog, specimen) -> None:
    """Test of https://github.com/khaeru/sdmx/issues/205."""
    with specimen("INSEE/gh-205.xml") as f:
        msg = sdmx.read_sdmx(f)
        assert isinstance(msg, sdmx.message.StructureMessage)

    # Messages were logged
    msg_template = "Could not resolve {cls}.concept_identity reference to ConceptScheme=FR1:CONCEPTS_INSEE(1.0) → Concept={id}"
    m1 = msg_template.format(cls="TimeDimension", id="TIME_PERIOD")
    m2 = msg_template.format(cls="PrimaryMeasure", id="OBS_VALUE")
    assert m1 in caplog.messages
    assert m2 in caplog.messages

    # Access the parsed DSD
    dsd = msg.structure["CNA-2014-PIB"]

    # Components have annotations with expected ID and text
    for component, text in (
        (dsd.dimensions.get("TIME_PERIOD"), m1),
        (dsd.measures.get("OBS_VALUE"), m2),
    ):
        a = component.annotations[0]
        assert "sdmx.reader.xml.v21-parse-error" == a.id
        assert text == str(a.text)


def test_gh_218(caplog, installed_schemas, specimen) -> None:
    """Test of https://github.com/khaeru/sdmx/pull/218."""
    with specimen("constructed/gh-218.xml") as f:
        # Specimen is XSD-valid
        validate_xml(f, installed_schemas)

        f.seek(0)

        # Specimen can be read
        msg = sdmx.read_sdmx(f)

    # The message sender has 1 contact, with all attributes populated
    assert isinstance(msg, sdmx.message.DataMessage) and msg.header.sender
    assert 1 == len(msg.header.sender.contact)
    contact = msg.header.sender.contact[0]
    assert contact.telephone is not None
    assert (
        1
        # Number of localizations of localizable attributes
        == len(contact.name.localizations)
        == len(contact.org_unit.localizations)
        == len(contact.responsibility.localizations)
        # Number of values of multi-value attributes
        == len(contact.email)
        == len(contact.fax)
        == len(contact.uri)
        == len(contact.x400)
    )


# Each entry is a tuple with 2 elements:
# 1. an instance of lxml.etree.Element to be parsed.
# 2. Either:
#   - A sdmx.model object, in which case the parsed element must match the object.
#   - A string, in which case parsing the element is expected to fail, raising an
#     exception matching the string.
ELEMENTS = [
    # xml._datetime()
    (  # with 5 decimal places
        E(qname("mes:Extracted"), "2020-08-18T00:14:31.59849+05:00"),
        datetime(2020, 8, 18, 0, 14, 31, 598490, tzinfo=timezone(timedelta(hours=5))),
    ),
    (  # with 7 decimal places
        E(qname("mes:Extracted"), "2020-08-18T01:02:03.4567891+00:00"),
        datetime(2020, 8, 18, 1, 2, 3, 456789, tzinfo=timezone.utc),
    ),
    (  # with "Z"
        E(qname("mes:Extracted"), "2020-08-18T00:14:31.59849Z"),
        datetime(2020, 8, 18, 0, 14, 31, 598490, tzinfo=timezone.utc),
    ),
    (  # with 7 decimal places AND "Z"; a message is logged on DEBUG (not checked)
        E(qname("mes:Extracted"), "2020-08-18T01:02:03.4567891Z"),
        datetime(2020, 8, 18, 1, 2, 3, 456789, tzinfo=timezone.utc),
    ),
    # xml._facet()
    (
        E(qname("str:TextFormat"), isSequence="False", startValue="3.4", endValue="1"),
        None,
    ),
    # …attribute names are munged; default textType is supplied
    (
        E(qname("str:EnumerationFormat"), minLength="1", maxLength="6"),
        Facet(
            type=FacetType(min_length=1, max_length=6),
            value_type=FacetValueType["string"],
        ),
    ),
    # …invalid attributes cause an exception
    (
        E(qname("str:TextFormat"), invalidFacetTypeAttr="foo"),
        re.compile("unexpected keyword argument 'invalid_facet_type_attr'"),
    ),
    # xml._key0: Create the necessary parent element to test the parsing of its child
    (E(qname("str:DataKeySet"), E(qname("str:Key")), isIncluded="True"), None),
    # xml._dks
    (E(qname("str:DataKeySet"), isIncluded="true"), None),
    # xml._pa
    (E(qname("str:ProvisionAgreement")), None),
]


@pytest.mark.parametrize(
    "elem, expected", ELEMENTS, ids=list(map(str, range(len(ELEMENTS))))
)
def test_parse_elem(elem, expected):
    """Test individual XML elements.

    This method allows unit-level testing of specific XML elements appearing in SDMX-ML
    messages. Add elements by extending the list passed to the parametrize() decorator.
    """
    # Convert to a file-like object compatible with convert()
    tmp = BytesIO(etree.tostring(elem))

    # Create a reader
    reader = Reader()

    if isinstance(expected, (str, re.Pattern)):
        # Parsing the element raises an exception
        with pytest.raises(XMLParseError, match=expected):
            reader.convert(tmp)
    else:
        # The element is parsed successfully
        result = reader.convert(tmp)

        if not result:
            stack = list(chain(*[s.values() for s in reader.stack.values()]))
            assert len(stack) == 1
            result = stack[0]

        if expected:
            # Expected value supplied
            assert expected == result


def test_availableconstraint_xml_response(specimen):
    """Test of https://github.com/khaeru/sdmx/pull/161"""

    with specimen("IMF_STA/availableconstraint_CPI.xml") as f:
        msg = sdmx.read_sdmx(f)

    res = sdmx.to_pandas(msg.constraint)
    assert len(res) == 1
    assert "CPI" in res.keys()
    assert len(res["CPI"]) == 1
    assert len(res["CPI"][0]) == 3

    assert isinstance(msg.constraint.CPI, ContentConstraint)

    assert list(res["CPI"][0].keys()) == ["COUNTRY", "FREQUENCY", "INDICATOR"]

    expected = pandas.Series(["111", "134"], name="COUNTRY")
    assert expected.equals(res["CPI"][0]["COUNTRY"])
    expected = pandas.Series(["A", "M", "Q"], name="FREQUENCY")
    assert expected.equals(res["CPI"][0]["FREQUENCY"])
    expected = pandas.Series(
        ["PCPIHA_IX", "PCPIHA_PC_CP_A_PT", "PCPI_PC_CP_A_PT"], name="INDICATOR"
    )
    assert expected.equals(res["CPI"][0]["INDICATOR"])



import pytest

from sdmx.format import Flag, MediaType, list_media_types


class TestMediaType:
    @pytest.fixture
    def mt(self):
        yield MediaType("foo", "xml", "3.0.0", Flag.meta | Flag.ts)

    def test_match(self, mt):
        other = "application/vnd.sdmx.foo+xml; version=3.0.1, bar=baz"
        assert mt.match(other) is True
        assert mt.match(other, strict=True) is False

    def test_flags(self, mt):
        assert mt.is_data is False
        assert mt.is_meta is True
        assert mt.is_structure_specific is False
        assert mt.is_time_series is True

    def test_repr(self, mt):
        assert "application/vnd.sdmx.foo+xml; version=3.0.0" == repr(mt)


@pytest.mark.parametrize("base, N", [("csv", 2), ("json", 8), ("xml", 14)])
def test_list_media_types(base, N):
    assert N == len(list_media_types(base=base))



"""Compare SDMX artefacts."""

import datetime
import enum
import logging
import textwrap
from collections import defaultdict
from collections.abc import Iterable
from copy import copy
from dataclasses import dataclass, fields, is_dataclass
from functools import singledispatch
from typing import Any, TypeVar, Union

import lxml.etree

from . import urn
from .model import internationalstring

log = logging.getLogger(__name__)


IGNORE_CONTEXT = {"Categorisation.artefact"}
VISITED: dict[tuple[int, int], set[int]] = defaultdict(set)


class Comparable:
    """Mix-in class for objects with a :meth:`.compare` method."""

    def compare(self, other, strict: bool = True, **options) -> bool:
        """Return :any:`True` if `self` is the same as `other`.

        `strict` and other `options` are used to construct an instance of
        :class:`Options`.
        """
        return compare(self, other, Options(self, strict=strict, **options))


@dataclass
class Options:
    """Options for a comparison."""

    #: Base object for a recursive comparison. Used internally for memoization/to
    #: improve performance.
    base: Any

    #: Objects compare equal even if :attr:`.IdentifiableArtefact.urn` is :any:`None`
    #: for either or both, so long as the URNs implied by their other attributes—that
    #: is, returned by :func:`sdmx.urn.make`—are the same.
    allow_implied_urn: bool = True

    #: Strict comparison: if :any:`True` (the default), then attributes and associated
    #: objects must compare exactly equal. If :any:`False`, then :any:`None` values on
    #: either side are permitted.
    strict: bool = True

    #: Level for log messages.
    log_level: int = logging.NOTSET

    #: Verbose comparison: continue comparing even after reaching a definitive
    #: :any:`False` result. If :attr:`log_level` is not set, :py:`verbose = True`
    #: implies :py:`log_level = logging.DEBUG`.
    verbose: bool = False

    _memo_key: tuple[int, int] = (0, 0)

    def __post_init__(self) -> None:
        # Create a key for memoization
        self._memo_key = (id(self.base), id(self))
        VISITED[self._memo_key].clear()

        # If no log level is given, set a default based on verbose
        if self.log_level == logging.NOTSET:
            self.log_level = {True: logging.DEBUG, False: logging.INFO}[self.verbose]

    def log(self, message: str, level: int = logging.INFO) -> None:
        """Log `message` on `level`.

        `level` must be at least :attr:`log_level`.
        """
        if level >= self.log_level:
            log.log(level, message)

    def visited(self, obj) -> bool:
        """Return :any:`True` if `obj` has already be compared."""
        if type(obj).__module__ == "builtins":
            return False

        entry = id(obj)

        if entry in VISITED[self._memo_key]:
            return True
        else:
            VISITED[self._memo_key].add(entry)
            return False


T = TypeVar("T", bound=object)


@singledispatch
def compare(left: object, right, opts: Options, context: str = "") -> bool:
    """Compare `left` to `right`.

    .. todo:: Reimplement as a subclass of :class:`.common.DispatchConverter`.
    """
    if is_dataclass(left):
        return compare_dataclass(left, right, opts, context)

    raise NotImplementedError(f"Compare {type(left)} {left!r} in {context}")


def compare_dataclass(left, right, opts: Options, context: str) -> bool:
    c = context or type(left).__name__

    result = right is not None
    for f in fields(left) if result else []:
        l_val, r_val = getattr(left, f.name), getattr(right, f.name)

        if opts.visited(l_val):
            continue  # Already compared to its counterpart

        c_sub = f"{c}.{f.name}"

        # Handle Options.allow_implied_urn
        if f.name == "urn" and not l_val is r_val is None and opts.allow_implied_urn:
            try:
                l_val = l_val or urn.make(left)
            except (AttributeError, ValueError):
                pass
            try:
                r_val = r_val or urn.make(right)
            except (AttributeError, ValueError):
                pass

        result_f = (
            l_val is r_val
            or compare(l_val, r_val, opts, c_sub)
            or c_sub in IGNORE_CONTEXT
        )

        result &= result_f

        if result_f is False:
            opts.log(f"Not identical: {c_sub}={shorten(l_val)} != {shorten(r_val)}")
            if not opts.verbose:
                break
        else:
            opts.log(f"{c_sub}={shorten(l_val)} == {shorten(r_val)}", logging.DEBUG)

    return result


# Built-in types


# TODO When dropping support for Python <=3.10, change to '@compare.register'
@compare.register(int)
@compare.register(str)
@compare.register(datetime.date)
def _eq(left: Union[int, str, datetime.date], right, opts, context=""):
    """Built-in types that must compare equal."""
    return left == right or (not opts.strict and right is None)


# TODO When dropping support for Python <=3.10, change to '@compare.register'
@compare.register(type(None))
@compare.register(bool)
@compare.register(float)
@compare.register(type)
@compare.register(enum.Enum)
def _is(left: Union[None, bool, float, type, enum.Enum], right, opts, context):
    """Built-in types that must compare identical."""
    return left is right or (not opts.strict and right is None or left is None)


@compare.register
def _(left: dict, right, opts, context=""):
    """Return :obj:`True` if `self` is the same as `other`.

    Two DictLike instances are identical if they contain the same set of keys, and
    corresponding values compare equal.
    """
    result = True

    l_keys = set(left.keys())
    r_keys = set(right.keys()) if hasattr(right, "keys") else set()
    if l_keys != r_keys:
        opts.log(
            f"Mismatched {type(left).__name__} keys: {shorten(sorted(l_keys))} "
            f"!= {shorten(sorted(r_keys))}"
        )
        result = False

    # Compare items pairwise
    for key in sorted(l_keys) if (result or opts.verbose and right is not None) else ():
        result &= compare(left[key], right.get(key, None), opts)
        if result is False and not opts.verbose:
            break

    return result


# TODO When dropping support for Python <=3.10, change to '@compare.register'
@compare.register(list)
@compare.register(set)
def _(left: Union[list, set], right, opts, context=""):
    if len(left) != len(right):
        opts.log(f"Mismatched length: {len(left)} != {len(right)}")
        return False

    try:
        l_values: Iterable = sorted(left)
        r_values: Iterable = sorted(right)
    except TypeError:
        l_values, r_values = left, right

    return all(
        compare(a, b, opts, f"{context}[{i}]")
        for i, (a, b) in enumerate(zip(l_values, r_values))
    )


# Types from upstream packages


@compare.register
def _(left: lxml.etree._Element, right, opts, context=""):
    try:
        r_val = copy(right)
        lxml.etree.cleanup_namespaces(r_val)
    except TypeError:
        return not opts.strict
    else:
        l_val = copy(left)
        lxml.etree.cleanup_namespaces(l_val)
        return lxml.etree.tostring(l_val) == lxml.etree.tostring(r_val)


# SDMX types


@compare.register
def _(left: internationalstring.InternationalString, right, opts, context=""):
    return compare(
        left.localizations, right.localizations, opts, f"{context}.localizations"
    )


def shorten(value: Any) -> str:
    """Return a shortened :func:`repr` of `value` for logging."""
    return textwrap.shorten(repr(value), 30, placeholder="…")



import logging
from collections.abc import Callable, Iterator
from dataclasses import Field, fields
from functools import lru_cache
from typing import Any, Iterable
from warnings import warn

log = logging.getLogger(__name__)


def compare(attr, a, b, strict: bool) -> bool:
    """Return :obj:`True` if ``a.attr`` == ``b.attr``.

    If strict is :obj:`False`, :obj:`None` is permissible as `a` or `b`; otherwise,

    .. deprecated:: 2.23.0

       Use :meth:`.Comparable.compare` or :func:`.compare` instead.
    """
    warn(
        "sdmx.util.compare(); use sdmx.compare instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return getattr(a, attr) == getattr(b, attr) or (
        not strict and None in (getattr(a, attr), getattr(b, attr))
    )


def only(iterator: Iterator) -> Any:
    """Return the only element of `iterator`, or :obj:`None`."""
    try:
        result = next(iterator)
        flag = object()
        assert flag is next(iterator, flag)
    except (StopIteration, AssertionError):
        return None  # 0 or ≥2 matches
    else:
        return result


@lru_cache()
def parse_content_type(value: str) -> tuple[str, dict[str, Any]]:
    """Return content type and parameters from `value`.

    Modified from :mod:`requests.util`.
    """
    # FIXME handle a value like text/html,application/xhtml+xml,application/xml;q=0.9,
    #       image/avif,image/webp,*/*;q=0.8
    tokens = value.split(";")
    content_type, params_raw = tokens[0].strip(), tokens[1:]
    params = {}
    to_strip = "\"' "

    for param in params_raw:
        k, *v = param.strip().split("=")

        if not k and not v:
            continue

        params[k.strip(to_strip).lower()] = v[0].strip(to_strip) if len(v) else True

    return content_type, params


def ucfirst(value: str) -> str:
    """Return `value` with its first character transformed to upper-case."""
    return value[0].upper() + value[1:]


_FIELDS_CACHE: dict[str, list[Field]] = dict()


def direct_fields(cls) -> Iterable[Field]:
    """Return the data class fields defined on `cls` or its class.

    This is like the ``__fields__`` attribute, but excludes the fields defined on any
    parent class(es).
    """
    # Key for `_FIELDS_CACHE`: the fully qualified name
    cls_name = f"{cls.__module__}.{cls.__name__}"

    try:
        return _FIELDS_CACHE[cls_name]
    except KeyError:
        parent_fields = set(fields(cls.mro()[1]))
        result = list(filter(lambda f: f not in parent_fields, fields(cls)))
        return _FIELDS_CACHE.setdefault(cls_name, result)


def preserve_dunders(cls: type, *names: str) -> Callable[[type], type]:
    """Copy dunder `names` from `cls` to a decorated class."""

    def decorator(other_cls: type) -> type:
        for name in map(lambda s: f"__{s}__", names):
            candidates: Iterator[Callable] = filter(
                None, map(lambda k: getattr(k, name), cls.__mro__)
            )
            setattr(other_cls, name, next(candidates))
        return other_cls

    return decorator



import re
from operator import itemgetter

import pytest

import sdmx
from sdmx import message
from sdmx.model import common, v21


@pytest.mark.parametrize(
    "cls",
    (message.Message, message.DataMessage, message.StructureMessage, message.Footer),
)
def test_compare(cls):
    A = cls()
    B = cls()

    assert A.compare(B) is True


class TestDataMessage:
    def test_update(self, caplog):
        dm = message.DataMessage()

        # A data structure with 3 dimensions
        dsd = v21.DataStructureDefinition()
        dim_foo = dsd.dimensions.getdefault("FOO")
        dsd.dimensions.getdefault("BAR")
        dsd.dimensions.getdefault("BAZ")

        # 0 data sets
        dm.update()
        assert None is dm.observation_dimension
        assert re.match("No DataSet in message", caplog.messages[-1])
        caplog.clear()

        # 1 data set, 0 observations
        ds0 = v21.DataSet(structured_by=dsd)
        dm.data.append(ds0)

        dm.update()
        assert None is dm.observation_dimension
        assert re.match(r"^Unable to .* 1 data set\(s\)", caplog.messages[-1])
        caplog.clear()

        # 1 data set, 1 observation, 3 dimensions at observation ('flat')
        obs0 = v21.Observation(
            dimension=dsd.make_key(common.Key, {"FOO": "f", "BAR": "b", "BAZ": "b"})
        )
        ds0.add_obs([obs0])

        dm.update()
        assert None is dm.observation_dimension
        assert re.match(r"^More than 1 dimension at observation", caplog.messages[-1])
        caplog.clear()

        # 1 data set, 1 observation, 1 dimension at observation
        obs0.dimension = dsd.make_key(common.Key, {"FOO": "f"})

        dm.update()
        assert dim_foo is dm.observation_dimension
        assert 0 == len(caplog.messages)

        # 2 data sets, 1 dimension at observation, but different
        ds1 = v21.DataSet(structured_by=dsd)
        obs1 = v21.Observation(dimension=dsd.make_key(common.Key, {"BAR": "b"}))
        ds1.add_obs([obs1])
        dm.data.append(ds1)

        dm.update()
        assert None is dm.observation_dimension
        assert re.match("Multiple data sets with different observ", caplog.messages[-1])


class TestStructureMessage:
    def test_add_contains_get(self) -> None:
        dsd = v21.DataStructureDefinition(id="foo")
        msg = message.StructureMessage()

        # add() stores the object
        msg.add(dsd)
        assert 1 == len(msg.structure)

        # __contains__() works
        assert dsd in msg

        # get() retrieves the object
        assert dsd is msg.get("foo")

        # add() with an object not collected in a StructureMessage raises TypeError
        item: "common.Item" = common.Item(id="bar")
        with pytest.raises(TypeError):
            msg.add(item)

        # __contains__() also raises TypeError
        with pytest.raises(TypeError):
            item in msg

        # get() with two objects of the same ID raises ValueError
        msg.add(v21.DataflowDefinition(id="foo"))
        with pytest.raises(ValueError):
            msg.get("foo")

    def test_dictlike_attribute_access(self) -> None:
        dsd = v21.DataStructureDefinition(id="foo")
        msg = message.StructureMessage()
        msg.add(dsd)

        # Attribute access works when added to the default, empty DictLike
        # NB This offends mypy because DictLike.__getattr__ is not defined explicitly
        assert msg.structure.foo is dsd  # type: ignore [attr-defined]

    @pytest.fixture
    def msg0(self) -> message.StructureMessage:
        msg = message.StructureMessage()

        # Add several objects
        for maintainer, id, version, with_urn in (
            ("FOO", "CL_BAR", "1.0", True),
            ("FOO", "CL_BAR", "1.1", False),
            ("BAZ", "CL_BAR", "1.0", True),
        ):
            cl: "common.Codelist" = common.Codelist(
                id=id, maintainer=common.Agency(id=maintainer), version=version
            )
            if with_urn:
                cl.urn = sdmx.urn.make(cl)
            msg.codelist[f"{maintainer}:{id}({version})"] = cl

        return msg

    def test_get(self, msg0: message.StructureMessage) -> None:
        # Get with ID only
        with pytest.raises(ValueError, match="ambiguous"):
            msg0.get("CL_BAR")

        # with partial URN: id + version
        assert msg0.get("CL_BAR(1.1)") is not None

        # with partial URN: maintainer + id
        assert msg0.get("BAZ:CL_BAR") is not None

        # with partial URN: class + maintainer + id
        assert msg0.get("Codelist=BAZ:CL_BAR") is not None

        # with partial URN: class + maintainer + id + version
        assert msg0.get("Codelist=FOO:CL_BAR(1.0)") is not None

        # with complete URN: class + maintainer + id + version
        assert (
            msg0.get("urn:sdmx:org.sdmx.infomodel.codelist.Codelist=FOO:CL_BAR(1.0)")
            is not None
        )

    def test_iter_objects(self) -> None:
        """:meth:`.iter_objects` can be used to iterate over all objects."""
        msg = message.StructureMessage()

        # Add several objects
        msg.add(common.Codelist(id="CL_FOO"))
        msg.add(common.Codelist(id="CL_BAR", is_external_reference=True))
        msg.add(common.Codelist(id="CL_BAZ"))

        msg.add(common.ConceptScheme(id="CS_FOO"))
        msg.add(common.ConceptScheme(id="CS_BAR", is_external_reference=True))
        msg.add(common.ConceptScheme(id="CS_BAZ"))

        # Method accepts external_reference arg; runs; returns an iterator
        result = msg.iter_objects()

        # All objects
        assert 6 == len(list(result))

        # All objects that are not external references are included
        assert {"CL_FOO", "CL_BAZ", "CS_FOO", "CS_BAZ"} == set(
            obj.id for obj in msg.iter_objects(external_reference=False)
        )

    def test_objects(self) -> None:
        """:meth:`.objects` can be used to access a collection according to class."""
        msg = message.StructureMessage()
        cl: "common.Codelist" = common.Codelist(id="foo")
        msg.add(cl)

        assert cl is msg.objects(common.Codelist)["foo"]

        with pytest.raises(TypeError):
            # TypeError is raised for classes not collected in this message
            msg.objects(v21.DataSet)


EXPECTED = [
    # Structure messages
    (
        "IPI-2010-A21-structure.xml",
        """<sdmx.StructureMessage>
  <Header>
    id: 'categorisation_1450864290565'
    prepared: '2015-12-23T09:51:30.565000+00:00'
    receiver: <Agency unknown>
    sender: <Agency FR1: Institut national de la statistique et des études économiques>
    source: fr: Banque de données macro-économiques
    test: False
  response: <Response [None]>
  Categorisation (1): CAT_IPI-2010_IPI-2010-A21
  CategoryScheme (1): CLASSEMENT_DATAFLOWS
  Codelist (7): CL_FREQ CL_NAF2_A21 CL_NATURE CL_UNIT CL_AREA CL_TIME_C...
  ConceptScheme (1): CONCEPTS_INSEE
  DataflowDefinition (1): IPI-2010-A21
  DataStructureDefinition (1): IPI-2010-A21""",
    ),
    (
        # This message shows the summarization feature: the DFD list is truncated
        "dataflow.xml",
        """<sdmx.StructureMessage>
  <Header>
    id: 'dataflow_ENQ-CONJ-TRES-IND-PERSP_1450865196042'
    prepared: '2015-12-23T10:06:36.042000+00:00'
    receiver: <Agency unknown>
    sender: <Agency FR1: Institut national de la statistique et des études économiques>
    source: fr: Banque de données macro-économiques
    test: False
  response: <Response [None]>
  DataflowDefinition (663): ACT-TRIM-ANC BPM6-CCAPITAL BPM6-CFINANCIER ...
  DataStructureDefinition (663): ACT-TRIM-ANC BPM6-CCAPITAL BPM6-CFINAN...""",
    ),
    # Data messages
    (
        "sg-xs.xml",
        """<sdmx.DataMessage>
  <Header>
    id: 'Generic'
    prepared: '2010-01-04T16:21:49+01:00'
    sender: <Agency ECB>
    source: """
        """
    test: False
  response: <Response [None]>
  DataSet (1)
  dataflow: <DataflowDefinition (missing id)>
  observation_dimension: <Dimension CURRENCY>""",
    ),
    (
        # This message has two DataSets:
        "action-delete.json",
        """<sdmx.DataMessage>
  <Header>
    id: '62b5f19d-f1c9-495d-8446-a3661ed24753'
    prepared: '2012-11-29T08:40:26+00:00'
    sender: <Agency ECB: European Central Bank>
    source: """
        """
    test: False
  response: <Response [None]>
  DataSet (2)
  dataflow: <DataflowDefinition (missing id)>
  observation_dimension: [<Dimension CURRENCY>]""",
    ),
]


@pytest.mark.parametrize(
    "pattern, expected", EXPECTED, ids=list(map(itemgetter(0), EXPECTED))
)
def test_message_repr(specimen, pattern, expected):
    import requests

    with specimen(pattern) as f:
        msg = sdmx.read_sdmx(f)

    # Attach a response object, as if the Message resulted from a requests query
    msg.response = requests.Response()

    if isinstance(expected, re.Pattern):
        assert expected.fullmatch(repr(msg))
    else:
        # __repr__() and __str__() give the same, expected result
        assert expected == repr(msg) == str(msg)



"""Convert :mod:`sdmx.message` and :mod:`.model` objects to :mod:`pandas` objects."""

from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from dataclasses import InitVar, dataclass, field
from itertools import chain, product, repeat
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any, Optional, Union, cast
from warnings import warn

import numpy as np
import pandas as pd

from sdmx import message, urn
from sdmx.dictlike import DictLike
from sdmx.format import csv
from sdmx.format.csv.common import Attributes, CSVFormatOptions, Labels, TimeFormat
from sdmx.format.csv.v2 import Keys
from sdmx.model import common, v21, v30
from sdmx.model.internationalstring import DEFAULT_LOCALE

from .common import DispatchConverter

if TYPE_CHECKING:
    from typing import TypedDict

    from pandas.core.dtypes.base import ExtensionDtype
    from pandas.tseries.offsets import DateOffset

    from sdmx.format.csv.common import CSVFormatOptions
    from sdmx.model.common import Item
    from sdmx.model.v21 import ContentConstraint

    class ToDatetimeKeywords(TypedDict, total=False):
        format: str

    KeyOrAttributeValue = Union["common.KeyValue", "common.AttributeValue"]


_HAS_PANDAS_2 = pd.__version__.split(".")[0] >= "2"

#: TODO Retrieve this info from the StructureMessage class.
ALL_CONTENTS = {
    "category_scheme",
    "codelist",
    "concept_scheme",
    "constraint",
    "dataflow",
    "structure",
    "organisation_scheme",
}


NO_VALUE = SimpleNamespace(value="")


class Column(ABC):
    """Representation of conversion of a column.

    .. todo:: Unify with :class:`.reader.csv.Handler`.
    """

    __slots__ = ("name", "id")

    #: Column name/header.
    name: str
    #: SDMX component ID.
    id: str

    @abstractmethod
    def __call__(self, *args, **kwargs) -> str: ...

    def __repr__(self) -> str:  # pragma: no cover
        return f"<{type(self).__name__} id={getattr(self, 'id', '—')!r} name={self.name!r}>"


class Fixed(Column):
    """Column with fixed value."""

    def __init__(self, name: str, value: str) -> None:
        self.name = name
        self._value = value

    def __call__(self) -> str:
        return self._value


class ComponentColumn(Column):
    """A column taking its header from a :class:`.Component`."""

    labels: Labels

    def __init__(self, component: "common.Component") -> None:
        self.id = component.id
        c_name = (
            str(component.concept_identity.name)
            if component.concept_identity
            else f"(Name of {component.id!r})"
        )
        # Set column header according to subclass `labels` attribute
        self.name = {
            Labels.id: self.id,
            Labels.both: f"{self.id}: {c_name}",
            Labels.name: c_name,
        }[self.labels]


class ComponentBoth(ComponentColumn):
    """:attr:`Labels.both` column."""

    labels = Labels.both

    def __call__(self, values: Mapping[str, "KeyOrAttributeValue"]) -> str:
        v = values[self.id].value
        return f"{v.id}: {v.name}" if isinstance(v, common.Code) else str(v)


class ComponentID(ComponentColumn):
    """:attr:`Labels.id` column."""

    labels = Labels.id

    def __call__(self, values: Mapping[str, "KeyOrAttributeValue"]) -> str:
        return str(values.get(self.id, NO_VALUE).value)


class ComponentName(ComponentColumn):
    """:attr:`Labels.name` column."""

    labels = Labels.name

    def __call__(self, values: Mapping[str, "KeyOrAttributeValue"]) -> str:
        v = values[self.id].value
        return str(v.name) if isinstance(v, common.Code) else v


class ColumnSpec:
    """Information about columns for conversion."""

    #: Initial columns.
    start: list[Fixed]
    #: Columns related to observation keys.
    key: list[Column]
    #: Column(s) related to observation measure(s).
    measure: list[Column]
    #: Columns related to observation-attached attributes.
    obs_attrib: list[Column]
    #: Final columns.
    end: list[Fixed]

    def __init__(
        self,
        pc: Optional["PandasConverter"] = None,
        ds: Optional["common.BaseDataSet"] = None,
    ) -> None:
        if pc is None or ds is None:
            return  # Empty/placeholder

        self.constraint = pc.constraint

        # Construct the short URN for the DFD
        dfd_urn = ""
        if pc._strict:
            if ds.described_by is None:
                raise ValueError(f"No associated data flow definition for {ds}")
            dfd_urn = urn.make(ds.described_by).partition("=")[2]

        # Either use a provided DSD, or construct one by inspecting the first
        # observation
        dsd = self._maybe_construct_dsd(
            pc._context.get(common.BaseDataStructureDefinition, None),
            ds.obs[0] if len(ds.obs) else common.BaseObservation(),
        )

        # Fixed columns
        self.start = {
            csv.common.CSVFormat: [],  # No specific SDMX-CSV format
            csv.v1.FORMAT: [Fixed("DATAFLOW", dfd_urn)],
            csv.v2.FORMAT: [
                Fixed("STRUCTURE", "dataflow"),
                Fixed("STRUCTURE_ID", dfd_urn),
                Fixed("ACTION", "I"),
            ],
        }[pc.format_options.format]

        # Identify classes for key and attribute columns
        classes = cast(
            list,
            {
                Labels.id: [ComponentID],
                Labels.both: [ComponentBoth],
                Labels.name: [ComponentID, ComponentName],
            }[pc.format_options.labels],
        )

        # Construct key columns: 1 or 2 columns for each dimension
        self.key = [cls(d) for d, cls in product(dsd.dimensions.components, classes)]

        # Measure columns
        _measures = dsd.measures.components
        if not _measures or pc.format_options.format is csv.v1.FORMAT:
            _measures = [
                v21.PrimaryMeasure(
                    id="OBS_VALUE",
                    concept_identity=common.Concept(
                        id="OBS_VALUE", name="Observation value"
                    ),
                )
            ]
        self.measure = [cls(m) for m, cls in product(_measures, classes)]

        # Attribute columns
        _attributes = dsd.attributes.components
        if pc.attributes is Attributes.none:
            # Omit attribute columns if so configured
            _attributes = []
            # Disable callback to extend self.obs_attrib
            setattr(self, "add_obs_attrib", lambda v: None)

        # Construct attribute columns: 1 or 2 columns for each attribute
        self.obs_attrib = [cls(a) for a, cls in product(_attributes, classes)]

        # Columns for attributes attached to the data set (SDMX 2.1 only)
        self.end = []
        if (pc.attributes & Attributes.dataset) and isinstance(ds, v21.DataSet):
            self.end.extend(Fixed(id, value) for id, value in ds.attrib.items())

    @staticmethod
    def _maybe_construct_dsd(
        dsd: Optional["common.BaseDataStructureDefinition"],
        obs: "common.BaseObservation",
    ) -> Union["v21.DataStructureDefinition", "v30.DataStructure"]:
        """If `dsd` is None, construct a DSD by inspection of `obs`."""
        if dsd is not None:
            return dsd

        result = v21.DataStructureDefinition()
        for dim_id in obs.key.order().values.keys():
            result.dimensions.getdefault(id=dim_id)

        for attr_id in obs.attrib:
            result.attributes.getdefault(id=attr_id)

        return result

    @property
    def assign(self) -> Mapping[str, str]:
        """Return values for :meth:`pandas.DataFrame.assign`."""
        return {c.name: c() for c in self.start + self.end}

    @property
    def full(self) -> list[str]:
        """Full list of column names."""
        return [c.name for c in self.start] + self.obs + [c.name for c in self.end]

    @property
    def obs(self) -> list[str]:
        """List of column names for observation data."""
        return [c.name for c in chain(self.key, self.measure, self.obs_attrib)]

    def add_obs_attrib(self, values: Iterable[str]) -> None:
        """Extend :attr:`obs_attrib` using `values`."""
        if missing := set(values) - set(c.id for c in self.obs_attrib):
            # Convert missing attributes to DataAttribute → Columns
            self.obs_attrib.extend(
                ComponentID(common.DataAttribute(id=id))
                for id in filter(missing.__contains__, values)
            )

    def convert_obs(self, obs: "common.BaseObservation") -> list:
        """Convert a single Observation to a data row.

        The items of the result correspond to the column names in :attr:`obs`.
        """
        key = obs.key
        if self.constraint and key not in self.constraint:
            # Emit an empty row to be dropped
            result: Iterable[Union[str, None]] = repeat(None, len(self.obs))
        else:
            # Observation values
            # FIXME Handled CodedObservationValue, similar to AttributeValue
            ov = None if obs.value is None else str(obs.value)

            # Combined attributes from observation, series-, and group keys
            avs = obs.attrib

            # Maybe update list of observation attributes
            self.add_obs_attrib(avs)

            # - Convert the observation Key using key Columns.
            # - Convert the value to Optional[str].
            # - Convert the attribute values using attribute Columns.
            result = chain(
                [c(key.values) for c in self.key],
                [ov] * len(self.measure),
                [c(avs) for c in self.obs_attrib],
            )
        return list(result)


@dataclass
class PandasConverter(DispatchConverter):
    """Convert SDMX messages and IM objects to :class:`pandas.DataFrame` and similar.

    PandasConverter implements a dispatch pattern according to the type of the object
    to be converted. The attributes/arguments to the class control the conversion
    behaviour and return types.
    """

    #: SDMX-CSV format options.
    format_options: "CSVFormatOptions" = field(default_factory=CSVFormatOptions)

    #: Attributes to include.
    attributes: Attributes = Attributes.none

    #: If given, only Observations included by the *constraint* are returned.
    constraint: Optional["ContentConstraint"] = None

    #: Datatype for observation values. If :any:`None`, data values remain
    #: :class:`object`/:class:`str`.
    dtype: Union[type["np.generic"], type["ExtensionDtype"], str, None] = np.float64

    #: Axis on which to place a time dimension. One of:
    #:
    #: - :py:`-1`: disabled.
    #: - :py:`0, "index"`: first/index axis.
    #: - :py:`1, "columns"`: second/columns axis.
    datetime_axis: Union[int, str] = -1

    #: Dimension to convert to :class:`pandas.DatetimeIndex`. A :class:`str` value is
    #: interpreted as a dimension ID.
    datetime_dimension: Optional["common.DimensionComponent"] = None

    #: Frequency for conversion to :class:`pandas.PeriodIndex`. A :class:`str` value is
    #: interpreted as one of the :ref:`pd:timeseries.period_aliases`.
    datetime_freq: Optional["DateOffset"] = None

    #: include : iterable of str or str, optional
    #:     One or more of the attributes of the StructureMessage ('category_scheme',
    #:     'codelist', etc.) to transform.
    include: set[str] = field(default_factory=lambda: set(ALL_CONTENTS))

    locale: str = DEFAULT_LOCALE

    #: :any:`True` to convert datetime.
    #:
    #: .. deprecated:: 2.23.0
    #:
    #:    Use :attr:`datetime_axis`, :attr:`datetime_dimension`, or
    #:    :attr:`datetime_freq`.
    datetime: InitVar = None

    #: Return type for :func:`.convert_dataset` and similar methods.
    #:
    #: .. deprecated:: 2.23.0
    #:
    #:    User code should instead explicitly modify the returned
    #:    :class:`~pandas.DataFrame` or :class:`~pandas.Series`.
    rtype: InitVar[str] = ""

    # Internal variables
    _columns: "ColumnSpec" = field(default_factory=ColumnSpec)

    # True if converting to SDMX-CSV
    _strict: bool = False

    # Columns to be set as index levels, then unstacked.
    _unstack: list[str] = field(default_factory=list)

    _context: dict[Union[str, type], Any] = field(
        default_factory=lambda: dict(compat=False)
    )

    def get_components(self, kind) -> list["common.Component"]:
        """Return an appropriate list of dimensions or attributes."""
        if ds := self._context.get(common.BaseDataSet, None):
            return getattr(ds.structured_by, kind).components
        elif dsd := self._context.get(
            common.BaseDataStructureDefinition
        ):  # pragma: no cover
            return getattr(dsd, kind).components
        else:  # pragma: no cover
            return []

    def handle_compat(self) -> None:
        """Analyse and alter settings for deprecated :py:`rtype=compat` argument."""
        if not self._context["compat"]:
            return

        # "Dimension at observation level" from the overall message
        obs_dim = self._context[message.DataMessage].observation_dimension

        if isinstance(obs_dim, common.TimeDimension):
            # Set datetime_dimension; convert_datetime() does the rest
            self.datetime_dimension = obs_dim
        elif isinstance(obs_dim, common.DimensionComponent):
            # Explicitly mark the other dimensions to be unstacked
            self._unstack = [
                d.id for d in self.get_components("dimensions") if d.id != obs_dim.id
            ]

    def handle_datetime(self, value: Any) -> None:
        """Handle alternate forms of :attr:`datetime`.

        If given, return a DataFrame with a :class:`~pandas.DatetimeIndex` or
        :class:`~pandas.PeriodIndex` as the index and all other dimensions as columns.
        Valid `datetime` values include:

        - :class:`bool`: if :obj:`True`, determine the time dimension automatically by
          detecting a :class:`~.TimeDimension`.
        - :class:`str`: ID of the time dimension.
        - :class:`~.Dimension`: the matching Dimension is the time dimension.
        - :class:`dict`: advanced behaviour. Keys may include:

           - **dim** (:class:`~.Dimension` or :class:`str`): the time dimension or its
             ID.
           - **axis** (`{0 or 'index', 1 or 'columns'}`): axis on which to place the time
             dimension (default: 0).
           - **freq** (:obj:`True` or :class:`str` or :class:`~.Dimension`): produce
             :class:`pandas.PeriodIndex`. If :class:`str`, the ID of a Dimension
             containing a frequency specification. If a Dimension, the specified
             dimension is used for the frequency specification.

             Any Dimension used for the frequency specification is does not appear in the
             returned DataFrame.
        """
        if value is None:
            return

        warn(
            f"datetime={value} argument of type {type(value)}. Instead, set other "
            "datetime_… fields directly.",
            DeprecationWarning,
            stacklevel=2,
        )

        if isinstance(value, (str, common.DimensionComponent)):
            self.datetime_dimension = value  # type: ignore [assignment]
        elif isinstance(value, dict):
            # Unpack a dict of 'advanced' arguments
            self.datetime_axis = value.pop("axis", self.datetime_axis)
            self.datetime_dimension = value.pop("dim", self.datetime_dimension)
            self.datetime_freq = value.pop("freq", self.datetime_freq)
            if len(value):
                raise ValueError(f"Unexpected datetime={tuple(sorted(value))!r}")
        elif isinstance(value, bool):
            self.datetime_axis = 0 if value else -1
        else:
            raise TypeError(f"PandasConverter(…, datetime={type(value)})")

    def __post_init__(self, datetime: Any, rtype: Optional[str]) -> None:
        """Transform and validate arguments."""
        # Raise on unsupported arguments
        if isinstance(
            self.format_options, csv.v2.FormatOptions
        ) and self.format_options.keys not in {Keys.none}:
            raise NotImplementedError(
                f"convert to SDMX-CSV with keys={self.format_options.keys}"
            )
        elif self.format_options.time_format not in {TimeFormat.original}:
            raise NotImplementedError(
                f"convert to SDMX-CSV with time_format={self.format_options.time_format}"
            )

        # Handle deprecated arguments
        self.handle_datetime(datetime)
        if rtype:
            warn(
                f"rtype={rtype!r} argument to to_pandas()/PandasConverter.",
                DeprecationWarning,
                stacklevel=2,
            )
            self._context["compat"] = rtype == "compat"

        # Convert other arguments to types expected by other code
        if isinstance(self.attributes, str):
            self.attributes = Attributes.parse(self.attributes)

        if isinstance(self.datetime_dimension, str):
            self.datetime_dimension = common.DimensionComponent(
                id=self.datetime_dimension
            )

        if isinstance(self.datetime_freq, str):
            try:
                # A frequency string recognized by pandas.PeriodDtype
                self.datetime_freq = pd.PeriodDtype(freq=self.datetime_freq).freq
            except ValueError:
                self.datetime_freq = common.Component(id=self.datetime_freq)

        if isinstance(self.include, str):
            self.include = set([self.include])
        # Silently discard invalid names
        self.include &= ALL_CONTENTS

        # Set private attributes
        self._strict = issubclass(
            self.format_options.format, (csv.v1.FORMAT, csv.v2.FORMAT)
        )


def to_pandas(obj, **kwargs):
    """Convert an SDMX `obj` to :mod:`pandas` object(s).

    `kwargs` can include any of the attributes of :class:`.PandasConverter`.

    Other parameters
    ----------------
    format_options :
        if not given, an instance of :class:`.CSVFormatOptions` is used as a default.
    labels :
        if given, the :attr:`.CSVFormatOptions.labels` attribute of the `format_options`
        keyword argument is replaced.
    time_format :
        if given, the :attr:`.CSVFormatOptions.time_format` attribute of the
        `format_options` keyword argument is replaced.

    .. versionchanged:: 1.0

       :func:`.to_pandas` handles all types of objects,
       replacing the earlier, separate ``data2pandas`` and ``structure2pd`` writers.

    .. versionchanged:: 2.23.0

       :func:`.to_pandas` is a thin wrapper for :class:`.PandasConverter`.
    """
    csv.common.kwargs_to_format_options(kwargs, csv.common.CSVFormatOptions)
    return PandasConverter(**kwargs).convert(obj)


# Functions for Python containers
@PandasConverter.register
def _list(c: "PandasConverter", obj: list):
    """Convert a :class:`list` of SDMX objects."""
    member_type = type(obj[0]) if len(obj) else object
    if issubclass(member_type, common.BaseDataSet) and 1 == len(obj):
        # Unpack a single data set
        return c.convert(obj[0])
    elif issubclass(member_type, common.BaseObservation):
        # Wrap a bare list of observations in DataSet
        return convert_dataset(c, v21.DataSet(obs=obj))
    elif issubclass(member_type, common.SeriesKey):
        # Return as pd.DataFrame instead of list
        return pd.DataFrame([c.convert(item) for item in obj])
    else:
        return [c.convert(item) for item in obj]


@PandasConverter.register
def _dict(c: "PandasConverter", obj: dict):
    """Convert mappings."""
    result = {k: c.convert(v) for k, v in obj.items()}

    result_type = set(type(v) for v in result.values())

    if result_type <= {pd.Series, pd.DataFrame}:
        if (
            len(set(map(lambda s: s.index.name, result.values()))) == 1
            and len(result) > 1
        ):
            # Can safely concatenate these to a pd.MultiIndex'd Series.
            return pd.concat(result)
        else:
            # The individual pd.Series are indexed by different dimensions; do not
            # concatenate
            return DictLike(result)
    elif result_type == {str}:
        return pd.Series(result)
    elif result_type < {dict, DictLike}:
        # Includes result_type == {}, i.e. no results
        return result
    else:  # pragma: no cover
        raise RuntimeError(f"Recursive conversion of {obj} returned {result_type}")


@PandasConverter.register
def _set(c: "PandasConverter", obj: set):
    """Convert :class:`set` recursively."""
    return {c.convert(o) for o in obj}


# Functions for message classes
@PandasConverter.register
def convert_datamessage(c: "PandasConverter", obj: message.DataMessage):
    """Convert :class:`.DataMessage`.

    Parameters
    ----------
    rtype : 'compat' or 'rows', optional
        Data type to return; default :data:`.DEFAULT_RTYPE`. See the
        :ref:`HOWTO <howto-rtype>`.
    kwargs :
        Passed to :func:`convert_dataset` for each data set.

    Returns
    -------
    :class:`pandas.Series` or :class:`pandas.DataFrame`
        if `obj` has only one data set.
    list of (:class:`pandas.Series` or :class:`pandas.DataFrame`)
        if `obj` has more than one data set.
    """
    # Update the context
    c._context[message.DataMessage] = obj
    # Use the specified structure of the message
    assert obj.dataflow
    c._context[common.BaseDataStructureDefinition] = obj.dataflow.structure
    # Handle the deprecated rtype="compat" argument
    c.handle_compat()

    # Convert list of data set objects
    result = c.convert(obj.data)

    c._context.pop(common.BaseDataStructureDefinition)
    c._context.pop(message.DataMessage)

    return result


@PandasConverter.register
def convert_structuremessage(c: "PandasConverter", obj: message.StructureMessage):
    """Convert :class:`.StructureMessage`.

    Returns
    -------
    .DictLike
        Keys are StructureMessage attributes; values are pandas objects.
    """
    attrs = sorted(c.include)
    result: DictLike[str, Union[pd.Series, pd.DataFrame]] = DictLike()
    for a in attrs:
        dl = c.convert(getattr(obj, a))
        if len(dl):
            # Only add non-empty elements
            result[a] = dl

    return result


# Functions for model classes


@PandasConverter.register
def _c(c: "PandasConverter", obj: common.Component):
    """Convert :class:`.Component`."""
    assert obj.concept_identity
    return str(obj.concept_identity.id)


@PandasConverter.register
def _cc(c: "PandasConverter", obj: v21.ContentConstraint):
    """Convert :class:`.ContentConstraint`."""
    return {i: c.convert(cr) for i, cr in enumerate(obj.data_content_region)}


@PandasConverter.register
def _cr(c: "PandasConverter", obj: common.CubeRegion):
    """Convert :class:`.CubeRegion`."""
    result: DictLike[str, pd.Series] = DictLike()
    for dim, ms in obj.member.items():
        result[dim.id] = pd.Series([c.convert(sv) for sv in ms.values], name=dim.id)
    return result


@PandasConverter.register
def _rp(c: "PandasConverter", obj: v21.RangePeriod):
    """Convert :class:`.RangePeriod`."""
    return f"{obj.start.period}–{obj.end.period}"


@PandasConverter.register
def convert_dataset(c: "PandasConverter", obj: common.BaseDataSet):
    """Convert :class:`~.DataSet`.

    See the :ref:`walkthrough <datetime>` for examples of using the `datetime` argument.

    Returns
    -------
    :class:`pandas.DataFrame`
        - if :attr:`~PandasConverter.attributes` is not ``''``, a data frame with one
          row per Observation, ``value`` as the first column, and additional columns
          for each attribute;
        - if `datetime` is given, various layouts as described above; or
        - if `_rtype` (passed from :func:`convert_datamessage`) is 'compat', various
          layouts as described in the :ref:`HOWTO <howto-rtype>`.
    :class:`pandas.Series` with :class:`pandas.MultiIndex`
        Otherwise.
    """
    c._context[common.BaseDataSet] = obj
    c._columns = ColumnSpec(pc=c, ds=obj)

    # - Apply convert_obs() to every obs → iterable of list.
    # - Create a pd.DataFrame.
    # - Drop empty rows (not in constraint).
    # - Set column names.
    # - Assign common values for all rows.
    # - Set column order.
    # - (Possibly) apply PandasConverter.dtype.
    # - (Possibly) convert certain columns to datetime.
    # - (Possibly) reshape.
    result = (
        pd.DataFrame(map(c._columns.convert_obs, obj.obs))
        .dropna(how="all")
        .set_axis(c._columns.obs, axis=1)  # NB This must come after DataFrame(map(…))
        .assign(**c._columns.assign)
        .pipe(_apply_dtype, c)
        .pipe(_convert_datetime, c)
        .pipe(_reshape, c)
        .pipe(_to_periodindex, c)
    )

    c._context.pop(common.BaseDataSet)

    return result


def _apply_dtype(df: "pd.DataFrame", c: "PandasConverter") -> "pd.DataFrame":
    """Apply `dtype` to 0 or more `columns`."""
    if c.dtype is None:
        return df

    # Create a mapping to apply `dtype` to multiple columns
    measure_cols = {c.name for c in c._columns.measure}
    dtypes = {col: c.dtype for col in measure_cols}

    try:
        return df.astype(dtypes)
    except ValueError:
        # Attempt to handle locales in which LC_NUMERIC.decimal_point is ","
        # TODO Make this more robust by inferring and changing locale settings
        assign_kw = {col: df[col].str.replace(",", ".") for col in measure_cols}
        return df.assign(**assign_kw).astype(dtypes)


def _convert_datetime(df: "pd.DataFrame", c: "PandasConverter") -> "pd.DataFrame":
    """Possibly convert a column to a pandas datetime dtype."""
    if c.datetime_dimension is c.datetime_freq is None and c.datetime_axis == -1:
        return df

    # Identify a time dimension
    dims = c.get_components("dimensions")
    try:
        dim = c.datetime_dimension or next(
            filter(lambda d: isinstance(d, common.TimeDimension), dims)
        )
    except StopIteration:
        raise ValueError(f"no TimeDimension in {dims}")

    # Record index columns to be unstacked
    c._unstack = c._unstack or list(map(str, filter(lambda d: d.id != dim.id, dims)))

    # Keyword args to pd.to_datetime(): only provide format= for pandas >=2.0.0
    dt_kw: "ToDatetimeKeywords" = dict(format="mixed") if _HAS_PANDAS_2 else {}

    # Convert the given column to a pandas datetime dtype
    return df.assign(**{dim.id: pd.to_datetime(df[dim.id], **dt_kw)})


def _ensure_multiindex(obj: Union[pd.Series, pd.DataFrame]):
    if not isinstance(obj.index, pd.MultiIndex):
        obj.index = pd.MultiIndex.from_product(
            [obj.index.to_list()], names=[obj.index.name]
        )
    return obj


def _reshape(
    df: "pd.DataFrame", c: "PandasConverter"
) -> Union[pd.Series, pd.DataFrame]:
    """Reshape `df` to provide expected return types."""

    if c._strict:
        # SDMX-CSV → no reshaping
        return df.reindex(columns=c._columns.full)

    # Set key columns as a pd.MultiIndex
    result = (
        df.set_index([col.name for col in c._columns.key])
        .pipe(_ensure_multiindex)
        .rename(columns={c.name: "value" for c in c._columns.measure})
    )

    # Single column for measure(s) + attribute(s) → return pd.Series
    if 1 == len(c._columns.obs) - len(c._columns.key):
        result = result.iloc[:, 0]

    # Unstack 1 or more index levels
    if c._unstack:
        result = result.unstack(c._unstack)

    return result


def _to_periodindex(obj: Union["pd.Series", "pd.DataFrame"], c: "PandasConverter"):
    """Convert a 1-D datetime index on `obj` to a PeriodIndex."""
    result = obj

    freq = c.datetime_freq

    # Convert to a PeriodIndex with a particular frequency
    if isinstance(freq, common.Component):
        # ID of a Dimension; Attribute; or column of `df`
        components = chain(
            c.get_components("dimensions"),
            c.get_components("attributes"),
            map(lambda id: common.Dimension(id=str(id)), result.columns.names),
        )
        try:
            component = next(filter(lambda c: c.id == freq.id, components))
        except StopIteration:
            raise ValueError(freq)

        if isinstance(component, common.Dimension):
            # Retrieve Dimension values from a pd.MultiIndex level
            level = component.id
            assert isinstance(result.columns, pd.MultiIndex)
            i = result.columns.names.index(level)
            values = set(result.columns.levels[i])
            # Remove the index level
            result.columns = result.columns.droplevel(i)
        elif isinstance(component, common.DataAttribute):  # pragma: no cover
            # Retrieve Attribute values from a column
            values = result[component.id].unique()

        if len(values) > 1:
            raise ValueError(
                f"cannot convert to PeriodIndex with non-unique freq={sorted(values)}"
            )

        # Store the unique value
        freq = values.pop()

    if freq is not None:
        assert isinstance(result.index, pd.DatetimeIndex)
        result.index = result.index.to_period(freq=freq)

    if c.datetime_axis in {1, "columns"}:
        result = result.transpose()

    return result


@PandasConverter.register
def _dd(c: "PandasConverter", obj: common.DimensionDescriptor):
    """Convert :class:`.DimensionDescriptor`.

    The collection of :attr:`.DimensionDescriptor.components` is converted.
    """
    return c.convert(obj.components)


@PandasConverter.register
def convert_itemscheme(c: "PandasConverter", obj: common.ItemScheme):
    """Convert :class:`.ItemScheme`.

    Parameters
    ----------
    locale : str, optional
        Locale for names to return.

    Returns
    -------
    pandas.Series or pandas.DataFrame
    """
    items = {}
    seen: set["Item"] = set()

    def add_item(item):
        """Recursive helper for adding items."""
        # Track seen items
        if item in seen:
            return
        else:
            seen.add(item)

        items[item.id] = dict(
            # Localized name
            name=item.name.localized_default(c.locale),
            # Parent ID
            parent=item.parent.id if isinstance(item.parent, item.__class__) else "",
        )

        # Add this item's children, recursively
        for child in item.child:
            add_item(child)

    for item in obj:
        add_item(item)

    # Convert to DataFrame
    result: Union[pd.DataFrame, pd.Series] = pd.DataFrame.from_dict(
        items,
        orient="index",
        dtype=object,  # type: ignore [arg-type]
    ).rename_axis(obj.id, axis="index")

    if len(result) and not result["parent"].str.len().any():
        # 'parent' column is empty; convert to pd.Series and rename
        result = result["name"].rename(obj.name.localized_default(c.locale))

    return result


@PandasConverter.register
def _mv(c: "PandasConverter", obj: common.BaseMemberValue):
    return obj.value


@PandasConverter.register
def _mds(c: "PandasConverter", obj: common.BaseMetadataSet):
    raise NotImplementedError(f"convert {type(obj).__name__} to pandas")


@PandasConverter.register
def _na(c: "PandasConverter", obj: common.NameableArtefact):
    """Fallback for NameableArtefact: only its name."""
    return str(obj.name)


@PandasConverter.register
def convert_serieskey(c: "PandasConverter", obj: common.SeriesKey):
    return {dim: kv.value for dim, kv in obj.order().values.items()}



import pandas as pd
import pandas.testing as pdt
import pytest

import sdmx
from sdmx import message
from sdmx.model import v21 as model
from sdmx.testing import MessageTest


class DataMessageTest(MessageTest):
    directory = "ECB_EXR"


class TestGenericFlatDataSet(DataMessageTest):
    filename = "ng-flat.xml"

    def test_msg_type(self, msg):
        assert isinstance(msg, message.DataMessage)

    def test_header_attributes(self, msg):
        # NB removed 2020-05-15. This is the <mes:Structure structureID="…"> attrib.
        #    SDMXCommon.xsd states: "The structureID attribute uniquely identifies the
        #    structure for the purpose of referencing it from the payload. This is only
        #    used in structure specific formats." It is not related to any DFD.
        # assert msg.dataflow.id == "STR1"

        # Maintained ID of the DataStructureDefinition is available
        assert msg.structure.id == "ECB_EXR_NG"
        assert msg.observation_dimension == model.AllDimensions

    def test_generic_obs(self, msg):
        data = msg.data[0]

        # No series
        assert len(data.series) == 0

        # Number of observations
        assert len(data.obs) == 12

        o0 = data.obs[0]

        assert isinstance(o0.key, model.Key)
        assert o0.key.FREQ == "M"
        assert o0.key.CURRENCY == "CHF"
        assert o0.value == "1.3413"

        assert o0.attrib.OBS_STATUS == "A"
        assert o0.attrib.DECIMALS == "4"

    def test_to_pandas(self, msg):
        # Single data series is converted to pd.Series
        data_series = sdmx.to_pandas(msg.data[0])
        assert isinstance(data_series, pd.Series)

        # When len(msg.data) is 1, the data series in a single Dataset are
        # unwrapped automatically
        assert len(msg.data) == 1
        data_series2 = sdmx.to_pandas(msg.data)  # NB no '[0]' index
        pdt.assert_series_equal(data_series, data_series2)


class TestGenericSeriesDataSet(DataMessageTest):
    filename = "ng-ts-gf.xml"

    def test_header_attributes(self, msg):
        # NB remove 2020-05-15; see above.
        # assert msg.dataflow.id == "STR1"
        assert msg.structure.id == "ECB_EXR_NG"
        assert msg.observation_dimension == "TIME_PERIOD"

    def test_generic_obs(self, msg):
        data = msg.data[0]

        # Number of observations in the entire dataset
        assert len(data.obs) == 12

        # Number of series in the dataset
        assert len(data.series) == 4

        # Access to series by index
        s3 = data.series[3]

        # Observations in series have .series_key with correct length & values
        assert isinstance(s3[0].series_key, model.Key)
        assert len(s3[0].series_key) == 5
        assert s3[0].series_key.CURRENCY == "USD"

        # Observations in series have attributes
        assert s3[0].attrib.DECIMALS == "4"

        # Number of observations in the series
        assert len(s3) == 3

        # Series observations can be reversed, and accessed by index
        o0 = list(reversed(s3))[2]

        # Series observations have expected value
        assert o0.dim == "2010-08"
        assert o0.value == "1.2894"

        assert o0.attrib.OBS_STATUS == "A"

    def test_pandas(self, msg):
        data = msg.data[0]

        series_keys = list(data.series.keys())

        # Number of series in dataframe
        assert len(series_keys) == 4

        # Convert the observations for one SeriesKey to a pd.Series
        s3_key = series_keys[3]
        s3 = sdmx.to_pandas(data.series[s3_key])
        assert isinstance(s3, pd.Series)

        # Test a particular value
        assert s3.iloc[0] == 1.2894

        # Length of index
        assert len(s3.index.names) == 6

        # Convert again, with attributes
        pd_data = sdmx.to_pandas(data, attributes="osgd")

        # Select one SeriesKey's data out of the DataFrame
        keys, levels = zip(*[(kv.value, kv.id) for kv in s3_key])
        s3 = pd_data.xs(keys, level=levels, drop_level=False)

        # Get the value of the first observation
        assert s3.iloc[0].value == 1.2894

        # Length of index
        assert len(s3.index.names) == 6

        # Number of attributes available
        assert len(set(s3.columns) - {"value"}) == 7

        # Access an attribute of the first value
        #
        # NB this uses the *pandas* attribute access shorthand, NOT DictLike:
        # "s3.iloc[0]" is a single row of s3, i.e. a pd.Series; and ".OBS_STATUS"
        # accesses the series element associated with that key in the index.
        assert s3.iloc[0].OBS_STATUS == "A"

        # Prior to khaeru/sdmx#243, the pandas object contains an AttributeValue object.
        # Now it contains only its value.
        with pytest.raises(AttributeError):
            assert s3.iloc[0].OBS_STATUS.value_for == "OBS_STATUS"

    def test_write2pandas(self, msg):
        df = sdmx.to_pandas(msg, attributes="")

        assert isinstance(df, pd.Series)

        assert df.shape == (12,)

        # with metadata
        df = sdmx.to_pandas(msg, attributes="osgd")
        df, mdf = df.iloc[:, 0], df.iloc[:, 1:]
        assert mdf.shape == (12, 7)
        assert mdf.iloc[1].OBS_STATUS == "A"


class TestGenericSeriesDataSet2(DataMessageTest):
    filename = "ng-ts.xml"

    def test_header_attributes(self, msg):
        # NB removed 2020-05-15; see above.
        # assert msg.dataflow.id == "STR1"
        assert msg.structure.id == "ECB_EXR_NG"

        # Observation dimension is 1 or more Dimensions
        assert msg.observation_dimension == "TIME_PERIOD"

    def test_generic_obs(self, msg):
        data = msg.data[0]

        assert len(data.obs) == 12

        assert len(data.series) == 4
        series_list = list(data.series.values())
        assert len(series_list) == 4
        s3 = series_list[3]

        # SeriesKey can be accessed by reference from each Observation in the
        # series
        assert isinstance(s3[0].series_key, model.Key)

        assert len(s3[0].series_key) == 5
        assert s3[0].key.CURRENCY == "USD"
        assert s3[0].attrib.DECIMALS == "4"
        obs_list = list(reversed(s3))
        assert len(obs_list) == 3
        o0 = obs_list[2]

        assert o0.dim == "2010-08"
        assert o0.value == "1.2894"

        assert o0.attrib.OBS_STATUS == "A"

    def test_dataframe(self, msg):
        df = sdmx.to_pandas(msg.data[0]).iloc[::-1]

        assert isinstance(df, pd.Series)

        assert df.shape == (12,)


class TestGenericSeriesData_SiblingGroup_TS(DataMessageTest):
    filename = "sg-ts.xml"

    def test_groups(self, msg):
        data = msg.data[0]

        # Data have expected number of groups and series
        assert len(data.group) == 4
        assert len(data.series) == 4

        # GroupKeys can be retrieved from keys of DataSet.group
        g2_key, g2 = list(data.group.items())[2]
        assert g2_key.CURRENCY == "JPY"
        assert g2[0].attrib.TITLE == ("ECB reference exchange rate, Japanese yen/Euro")

        # Check group attributes of a series
        s = list(data.series)[0]
        g_attrib = s.group_attrib
        assert len(g_attrib) == 1


class TestGenericSeriesData_RateGroup_TS(DataMessageTest):
    filename = "rg-ts.xml"

    def test_groups(self, msg):
        data = msg.data[0]

        assert len(data.group) == 5
        assert len(data.series) == 4

        # .group is DictLike; retrieve the key and obs separately
        g2_key, g2 = list(data.group.items())[2]
        assert g2_key.CURRENCY == "GBP"
        assert g2_key.attrib.TITLE == (
            "ECB reference exchange rate, U.K. Pound sterling /Euro"
        )
        # Check group attributes of a series
        s = list(data.series)[0]
        g_attrib = s.group_attrib
        assert len(g_attrib) == 5

    def test_footer(self, specimen):
        with specimen("footer.xml") as f:
            f = sdmx.read_sdmx(f).footer
        assert f.code == 413
        assert f.severity == "Infomation"
        assert str(f.text[1]).startswith("http")



import pytest

from sdmx.session import Session

from . import has_requests_cache


class TestSession:
    def test_init0(self, tmp_path):
        # requests_cache keyword argument, with invalid value
        # - TypeError if requests_cache is not installed
        # - ValueError from requests_cache if installed
        with pytest.raises(ValueError if has_requests_cache else TypeError):
            Session(backend="not_a_backend")

        # Keyword argument for a requests-cache backend, like fast_save, are accepted
        Session(
            cache_name=tmp_path.joinpath("TestSession_test_init0"),
            backend="sqlite",
            fast_save=True,
        )

        # Not an argument handled by sdmx1 or by requests_cache —no exception raised
        Session(foo="bar")


@pytest.mark.network
@pytest.mark.xfail(condition=not has_requests_cache, reason="Requires cache")
def test_init_cache(tmp_path):
    # Instantiate a REST object with cache
    cache_name = tmp_path.joinpath("test_init_cache")
    s = Session(cache_name=str(cache_name), backend="sqlite")

    # Get a resource
    s.get("https://registry.sdmx.org/ws/rest/dataflow")

    # Test for existence of cache file
    assert cache_name.with_suffix(".sqlite").exists()



from . import Source as BaseSource


class Source(BaseSource):
    _id = "WB_WDI"

    def modify_request_args(self, kwargs):
        """World Bank's agency ID."""
        super().modify_request_args(kwargs)

        if kwargs["resource_type"] == "categoryscheme":
            # Service does not respond to requests for "WB" category schemes
            kwargs["provider"] = "all"
        elif kwargs["resource_type"] != "data":
            # Provider's own ID differs from its ID in this package
            kwargs.setdefault("provider", "WB")

        try:
            if isinstance(kwargs["key"], str):
                # Data queries fail without a trailing slash
                # TODO improve the hook integration with Client.get so this can be done
                #      after the query URL is prepared
                kwargs["key"] += "/"
        except KeyError:
            pass



import logging

import pytest

from sdmx.message import StructureMessage
from sdmx.writer.protobuf import write as to_protobuf


@pytest.mark.xfail(raises=RuntimeError, reason="sdmx.format.protobuf_pb2 missing")
def test_codelist(caplog, codelist):
    msg = StructureMessage()
    msg.codelist[codelist.id] = codelist

    caplog.set_level(logging.ERROR)

    result = to_protobuf(msg)

    print(result)

    # No errors logged
    assert len(caplog.messages) == 0



import json

from sdmx.testing.report import main


def test_report_main(tmp_path):
    # Example input data
    with open(tmp_path.joinpath("TEST.json"), "w") as f:
        json.dump({"TEST": {"foo": "pass"}}, f)

    # Function runs
    main(tmp_path)

    # Output files are generated
    assert tmp_path.joinpath("all-data.json").exists()
    assert tmp_path.joinpath("index.html").exists()



"""Classes for SDMX messages.

:class:`Message` and related classes are not defined in the SDMX
:doc:`information model <implementation>`, but in the :ref:`SDMX-ML standard <formats>`.

:mod:`sdmx` also uses :class:`DataMessage` to encapsulate SDMX-JSON data returned by
data sources.
"""

import logging
import re
from dataclasses import dataclass, field, fields
from datetime import datetime
from itertools import chain
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    Generator,
    Optional,
    Text,
    Union,
    get_args,
)

import sdmx.urn
from sdmx import model
from sdmx.compare import Comparable
from sdmx.dictlike import DictLike, summarize_dictlike
from sdmx.dictlike import DictLikeDescriptor as DLD
from sdmx.format import Version
from sdmx.model import common, v21, v30
from sdmx.model.internationalstring import (
    InternationalString,
    InternationalStringDescriptor,
)
from sdmx.util import direct_fields

if TYPE_CHECKING:
    import requests

log = logging.getLogger(__name__)


def _summarize(obj, include: Optional[list[str]] = None):
    """Helper method for __repr__ on Header and Message (sub)classes."""
    import requests

    include = include or list(map(attrgetter("name"), fields(obj)))
    for name in include:
        attr = getattr(obj, name)
        if attr is None:
            continue
        elif isinstance(attr, datetime):
            s_attr = repr(attr.isoformat())
        elif isinstance(attr, requests.Response):
            # Don't use repr(), which displays the entire response body
            s_attr = str(attr)
        else:
            s_attr = repr(attr)

        yield f"{name}: {s_attr}"


@dataclass
class Header:
    """Header of an SDMX-ML message.

    SDMX-JSON messages do not have headers.
    """

    #: (optional) Error code for the message.
    error: Optional[Text] = None
    #: Date and time at which the data was extracted.
    extracted: Optional[datetime] = None
    #: Identifier for the message.
    id: Optional[Text] = None
    #: Date and time at which the message was generated.
    prepared: Optional[datetime] = None
    #: Start of the time period covered by a :class:`.DataMessage`.
    reporting_begin: Optional[datetime] = None
    #: End of the time period covered by a :class:`.DataMessage`.
    reporting_end: Optional[datetime] = None
    #: Intended recipient of the message, e.g. the user's name for an
    #: authenticated service.
    receiver: Optional[model.Agency] = None
    #: The :class:`.Agency` associated with the data :class:`~.source.Source`.
    sender: Optional[model.Agency] = None
    #:
    source: InternationalStringDescriptor = InternationalStringDescriptor()
    #:
    test: bool = False

    def __repr__(self):
        """String representation."""
        lines = ["<Header>"]
        lines.extend(_summarize(self))
        return "\n  ".join(lines)


@dataclass
class Footer(Comparable):
    """Footer of an SDMX-ML message.

    SDMX-JSON messages do not have footers.
    """

    #:
    severity: Optional[str] = None
    #: The body text of the Footer contains zero or more blocks of text.
    text: list[model.InternationalString] = field(default_factory=list)
    #:
    code: Optional[int] = None

    def __post_init__(self):
        # Convert non-IS members to IS
        self.text = [
            t if isinstance(t, InternationalString) else InternationalString(t)
            for t in self.text
        ]


@dataclass
class Message(Comparable):
    #: SDMX version.
    version: Version = Version["2.1"]

    #: :class:`Header` instance.
    header: Header = field(default_factory=Header)
    #: (optional) :class:`Footer` instance.
    footer: Optional[Footer] = None
    #: :class:`requests.Response` instance for the response to the HTTP request that
    #: returned the Message. This is not part of the SDMX standard.
    response: Optional["requests.Response"] = None

    def __str__(self):
        return repr(self)

    def __repr__(self):
        """String representation."""
        lines = [
            f"<sdmx.{self.__class__.__name__}>",
            repr(self.header).replace("\n", "\n  "),
        ]
        lines.extend(_summarize(self, ["footer", "response"]))
        return "\n  ".join(lines)


class ErrorMessage(Message):
    pass


@dataclass
class StructureMessage(Message):
    """SDMX StructureMessage."""

    #: Collection of :class:`.Categorisation`.
    categorisation: DLD[str, model.Categorisation] = DLD()
    #: Collection of :class:`.CategoryScheme`.
    category_scheme: DLD[str, model.CategoryScheme] = DLD()
    #: Collection of :class:`.Codelist`.
    codelist: DLD[str, model.Codelist] = DLD()
    #: Collection of :class:`.ConceptScheme`.
    concept_scheme: DLD[str, model.ConceptScheme] = DLD()
    #: Collection of :class:`.ContentConstraint`.
    constraint: DLD[str, model.BaseConstraint] = DLD()
    #: Collection of :class:`.CustomTypeScheme`.
    custom_type_scheme: DLD[str, model.CustomTypeScheme] = DLD()
    #: Collection of :class:`Dataflow(Definition) <.BaseDataflow>`.
    dataflow: DLD[str, model.BaseDataflow] = DLD()
    #: Collection of :class:`.HierarchicalCodelist`.
    hierarchical_codelist: DLD[str, v21.HierarchicalCodelist] = DLD()
    #: Collection of :class:`.v30.Hierarchy`.
    hierarchy: DLD[str, v30.Hierarchy] = DLD()
    #: Collection of :class:`Metadataflow(Definition) <.BaseMetadataflow>`.
    metadataflow: DLD[str, model.BaseMetadataflow] = DLD()
    #: Collection of :class:`MetadataStructureDefinition
    #: <.BaseMetadataStructureDefinition>`.
    metadatastructure: DLD[str, model.BaseMetadataStructureDefinition] = DLD()
    #: Collection of :class:`.NamePersonalisationScheme`.
    name_personalisation_scheme: DLD[str, model.NamePersonalisationScheme] = DLD()
    #: Collection of :class:`.OrganisationScheme`.
    organisation_scheme: DLD[str, model.OrganisationScheme] = DLD()
    #: Collection of :class:`.ProvisionAgreement`.
    provisionagreement: DLD[str, model.ProvisionAgreement] = DLD()
    #: Collection of :class:`.RulesetScheme`.
    ruleset_scheme: DLD[str, model.RulesetScheme] = DLD()
    #: Collection of :class:`DataStructureDefinition <.BaseDataStructureDefinition>`.
    structure: DLD[str, model.BaseDataStructureDefinition] = DLD()
    #: Collection of :class:`.StructureSet`.
    structureset: DLD[str, v21.StructureSet] = DLD()
    #: Collection of :class:`.TransformationScheme`.
    transformation_scheme: DLD[str, model.TransformationScheme] = DLD()
    #: Collection of :class:`.UserDefinedOperatorScheme`.
    user_defined_operator_scheme: DLD[str, model.UserDefinedOperatorScheme] = DLD()
    #: Collection of :class:`.ValueList` (SDMX 3.0 only).
    valuelist: DLD[str, v30.ValueList] = DLD()
    #: Collection of :class:`.VTLMappingScheme`.
    vtl_mapping_scheme: DLD[str, model.VTLMappingScheme] = DLD()

    def __post_init__(self):
        # Construct a list referencing all of the collections
        self._collections = [
            getattr(self, f.name) for f in direct_fields(self.__class__)
        ]

    def add(self, obj: model.IdentifiableArtefact):
        """Add `obj` to the StructureMessage."""
        for f in direct_fields(self.__class__):
            # NB for some reason mypy complains here, but not in __contains__(), below
            if isinstance(obj, get_args(f.type)[1]):
                getattr(self, f.name)[obj.id] = obj
                return
        raise TypeError(type(obj))

    def get(
        self, obj_or_id: Union[str, model.IdentifiableArtefact]
    ) -> Optional[model.IdentifiableArtefact]:
        """Retrieve `obj_or_id` from the StructureMessage.

        Parameters
        ----------
        obj_or_id : str or .IdentifiableArtefact
            - if an IdentifiableArtefact, return an object of the same class and
              :attr:`~.IdentifiableArtefact.id`.
            - if :class:`str`, this may be:

              - An exact match for some :attr:`.IdentifiableArtefact.id`.
              - Part of an SDMX :class:`URN`, for instance "FOO(1.2.3)", "MAINT:FOO", or
                "MAINT:FOO(1.2.3)".
              - A full SDMX :class:`URN`.

        Returns
        -------
        .IdentifiableArtefact
            with the given ID and possibly class, or :any:`None` if there is no match.

        Raises
        ------
        ValueError
            if there are ≥2 objects with the same `obj_or_id`; for instance, two objects
            of different classes, or two objects of the same class with different
            :attr:`~.MaintainableArtefact.maintainer` or
            :attr:`~.VersionableArtefact.version`.
        """
        id_ = (
            obj_or_id.id
            if isinstance(obj_or_id, model.IdentifiableArtefact)
            else obj_or_id
        )

        # Regular expression for matching object URNs
        try:
            # - Handle `id_` as if it is a partial or complete URN.
            # - Convert to a regular expression pattern.
            # - If the version is not given, match any version.
            urn_expr = re.escape(str(sdmx.urn.URN(sdmx.urn.expand(id_)))).replace(
                r"\(None\)", r"\([^\)]*\)"
            )
        except ValueError:
            # `id_` is not a (partial) URN. Match it `id_` immediately after one of the
            # separator characters
            urn_expr = rf"[=:\.]{re.escape(id_)}"

        urn_pat = re.compile(urn_expr)  # Compile re.Pattern object once

        candidates: list[model.IdentifiableArtefact] = []
        for key, obj in chain(*[c.items() for c in self._collections]):
            # Obtain a matchable string with the URN of `obj`
            try:
                urn = obj.urn or sdmx.urn.make(obj)  # Existing or constructed URN
            except ValueError:
                urn = ""  # No existing URN and unable to construct one
            if id_ in (key, obj.id) or urn_pat.search(urn):
                candidates.append(obj)

        if len(candidates) > 1:
            raise ValueError(f"ambiguous; {repr(obj_or_id)} matches {repr(candidates)}")

        return candidates[0] if len(candidates) == 1 else None

    def iter_collections(self) -> Generator[tuple[str, type], None, None]:
        """Iterate over collections."""
        for f in direct_fields(self.__class__):
            yield f.name, get_args(f.type)[1]

    def iter_objects(
        self, external_reference: bool = True
    ) -> Generator[common.MaintainableArtefact, None, None]:
        """Iterate over all objects in the message."""
        for _, cls in self.iter_collections():
            for obj in self.objects(cls).values():
                if not external_reference and obj.is_external_reference:
                    continue
                yield obj

    def objects(self, cls):
        """Get a reference to the attribute for objects of type `cls`.

        For example, if `cls` is the class :class:`DataStructureDefinition` (not an
        instance), return a reference to :attr:`structure`.
        """
        for f in direct_fields(self.__class__):
            if issubclass(cls, get_args(f.type)[1]):
                return getattr(self, f.name)
        raise TypeError(cls)

    def __contains__(self, item):
        """Return :obj:`True` if `item` is in the StructureMessage."""
        for f in direct_fields(self.__class__):
            if isinstance(item, get_args(f.type)[1]):
                return item in getattr(self, f.name).values()
        raise TypeError(f"StructureMessage has no collection of {type(item)}")

    def __repr__(self):
        """String representation."""
        lines = [super().__repr__()]

        # StructureMessage contents
        for attr in self.__dict__.values():
            if isinstance(attr, DictLike) and attr:
                lines.append(summarize_dictlike(attr))

        return "\n  ".join(lines)


@dataclass
class DataMessage(Message):
    """SDMX Data Message.

    .. note:: A DataMessage may contain zero or more :class:`.DataSet`, so :attr:`data`
       is a list. To retrieve the first (and possibly only) data set in the message,
       access the first element of the list: :py:`msg.data[0]`.
    """

    #: :class:`list` of :class:`.DataSet`.
    data: list[model.BaseDataSet] = field(default_factory=list)
    #: :class:`.DataflowDefinition` that contains the data.
    dataflow: Optional[model.BaseDataflow] = None
    #: The "dimension at observation level".
    observation_dimension: Optional[
        Union[
            model._AllDimensions,
            model.DimensionComponent,
            list[model.DimensionComponent],
        ]
    ] = None

    def __post_init__(self):
        if self.dataflow is None:
            # Create a default of the appropriate class
            self.dataflow = {
                Version["2.1"]: v21.DataflowDefinition,
                Version["3.0.0"]: v30.Dataflow,
            }[self.version]()

    # Convenience access
    @property
    def structure(self):
        """DataStructureDefinition used in the :attr:`dataflow`."""
        return self.dataflow.structure

    @property
    def structure_type(self) -> type[common.Structure]:
        """:class:`.Structure` subtype describing the contained (meta)data."""
        return {
            Version["2.1"]: v21.DataStructureDefinition,
            Version["3.0.0"]: v30.DataStructureDefinition,
        }[self.version]

    def __repr__(self):
        """String representation."""
        lines = [super().__repr__()]

        # DataMessage contents
        if self.data:
            lines.append("DataSet ({})".format(len(self.data)))
        lines.extend(_summarize(self, ("dataflow", "observation_dimension")))

        return "\n  ".join(lines)

    def update(self) -> None:
        """Update :attr:`.observation_dimension`.

        The observation dimensions (or dimension observation) is determined
        automatically if:

        1. There is at least 1 :class:`DataSet <.BaseDataSet>` in the message.
        2. For at least 1 data set:

           - :attr:`~.BaseDataSet.structured_by` is defined.
           - There is at least 1 :class:`.Observation` in the data set. (:meth:`.update`
             checks only the first observation.)
           - The :attr:`.Observation.dimension` is a :class:`.Key` referring to exactly
             1 dimension.

        3. The dimension indicated by (2) is the same for all DataSets in the message.

        If not all these conditions are met, messages are logged with level DEBUG, and
        :attr:`.observation_dimension` is set to :any:`None`.

        .. note:: :meth:`.update` is not automatically called when data sets are added
           to or removed from :attr:`.data`. User code **should** call :meth:`.update`
           to reflect such changes.
        """
        if not self.data:
            log.debug("No DataSet in message")
            self.observation_dimension = None
            return

        dims = set()
        for ds in self.data:
            try:
                assert ds.structured_by

                # Use the first observation
                assert len(ds.obs)
                o0 = ds.obs[0]
                assert o0.dimension

                # Identify the dimensions specified per-observation
                d_a_o = tuple(o0.dimension.values.keys())

                if 1 == len(d_a_o):
                    # Single dimension-at-observation
                    # Record as an attribute of the DataMessage
                    dims.add(ds.structured_by.dimensions.get(d_a_o[0]))
                else:
                    dims.add(d_a_o)
            except AssertionError:
                continue

        if len(dims) == 1 and not all(isinstance(d, tuple) for d in dims):
            self.observation_dimension = dims.pop()
        else:
            if len(dims) == 1:
                log.debug(f"More than 1 dimension at observation level: {dims.pop()}")
            elif len(dims) > 1:
                log.debug(
                    f"Multiple data sets with different observation dimension: {dims}"
                )
            elif not dims:
                log.debug(
                    f"Unable to determine observation dimension for {len(self.data)} "
                    "data set(s). Data set(s) may lack structure reference or "
                    "observations."
                )
            self.observation_dimension = None


@dataclass
class MetadataMessage(DataMessage):
    """SDMX Metadata Message."""

    @property
    def structure_type(self) -> type[common.Structure]:
        return {
            Version["2.1"]: v21.MetadataStructureDefinition,
            Version["3.0.0"]: v30.MetadataStructureDefinition,
        }[self.version]


class RegistryInterface(Message):
    """Common base class for registry interface messages."""


@dataclass
class SubmitStructureResponse(RegistryInterface):
    """SDMX SubmitStructureResponse."""

    result: list[common.SubmissionResult] = field(default_factory=list)



# TODO tidy these tests to use fixtures/methods from sdmx.tests
from collections import OrderedDict
from pathlib import Path

import pytest

import sdmx

BASE_PATH = Path(".") / "INSEE"

DATAFLOW_FP = "dataflow.xml"

DATASETS = {
    "IPI-2010-A21": {
        "data-fp": "IPI-2010-A21.xml",
        "datastructure-fp": "IPI-2010-A21-structure.xml",
        "series_count": 20,
    },
    "CNA-2010-CONSO-SI-A17": {
        "data-fp": "CNA-2010-CONSO-SI-A17.xml",
        "datastructure-fp": "CNA-2010-CONSO-SI-A17-structure.xml",
        "series_count": 1,
    },
}

SERIES = {"UNEMPLOYMENT_CAT_A_B_C": {"data-fp": "bug-series-freq-data.xml"}}


class TestINSEE:
    @pytest.fixture
    def base_path(self, test_data_path):
        yield test_data_path / "INSEE"

    def test_load_dataset(self, base_path):
        dataset_code = "IPI-2010-A21"

        # Load all dataflows
        dataflows_response = sdmx.read_sdmx(base_path / DATAFLOW_FP)
        dataflows = dataflows_response.dataflow

        assert len(dataflows) == 663
        assert dataset_code in dataflows

        # Load datastructure for current dataset_code
        fp_datastructure = base_path / DATASETS[dataset_code]["datastructure-fp"]
        datastructure_response = sdmx.read_sdmx(fp_datastructure)
        assert dataset_code in datastructure_response.dataflow
        dsd = datastructure_response.dataflow[dataset_code].structure

        # Verify dimensions list
        dimensions = OrderedDict(
            [dim.id, dim]
            for dim in dsd.dimensions
            if dim.id not in ["TIME", "TIME_PERIOD"]
        )
        dim_keys = list(dimensions.keys())
        assert dim_keys == ["FREQ", "PRODUIT", "NATURE"]

        # Load datas for the current dataset
        fp_data = base_path / DATASETS[dataset_code]["data-fp"]
        data = sdmx.read_sdmx(fp_data)

        # Verify series count and values
        series = data.data[0].series
        series_count = len(series)
        assert series_count == DATASETS[dataset_code]["series_count"]

        first_series = series[0]
        observations = first_series

        first_obs = observations[0]
        last_obs = observations[-1]

        assert first_obs.dim == "2015-10"
        assert first_obs.value == "105.61"

        assert last_obs.dim == "1990-01"
        assert last_obs.value == "139.22"

    def test_fixe_key_names(self, base_path):
        """Verify key or attribute contains '-' in name."""
        dataset_code = "CNA-2010-CONSO-SI-A17"

        fp_datastructure = base_path / DATASETS[dataset_code]["datastructure-fp"]
        datastructure_response = sdmx.read_sdmx(fp_datastructure)
        assert dataset_code in datastructure_response.dataflow
        dsd = datastructure_response.dataflow[dataset_code].structure

        dimensions = OrderedDict(
            [dim.id, dim]
            for dim in dsd.dimensions
            if dim.id not in ["TIME", "TIME_PERIOD"]
        )
        dim_keys = list(dimensions.keys())
        assert dim_keys == ["SECT-INST", "OPERATION", "PRODUIT", "PRIX"]

        fp_data = base_path / DATASETS[dataset_code]["data-fp"]
        data = sdmx.read_sdmx(fp_data)
        series = data.data[0].series
        series_key = list(series.keys())[0]

        assert list(series_key.values.keys()) == [
            "SECT-INST",
            "OPERATION",
            "PRODUIT",
            "PRIX",
        ]

        assert list(series_key.attrib.keys()) == [
            "FREQ",
            "IDBANK",
            "TITLE",
            "LAST_UPDATE",
            "UNIT_MEASURE",
            "UNIT_MULT",
            "REF_AREA",
            "DECIMALS",
            "BASE_PER",
            "TIME_PER_COLLECT",
        ]

    def test_freq_in_series_attribute(self, base_path):
        # Test that we don't have regression on Issues #39 and #41
        # INSEE time series provide the FREQ value as attribute on the series
        # instead of a dimension. This caused a runtime error when writing as
        # pandas dataframe.
        data_response = sdmx.read_sdmx(
            base_path / SERIES["UNEMPLOYMENT_CAT_A_B_C"]["data-fp"]
        )
        sdmx.to_pandas(data_response)



"""Common code for describing SDMX data formats."""

from typing import ClassVar


class Format:
    """Information about a SDMX data/file format.

    Any concrete subclass corresponds to a specific version of a data/file format
    defined in a specific version of the SDMX standards.
    """

    #: Format version.
    version: ClassVar[str]

    #: Preferred file name suffix.
    suffix: ClassVar[str]


class FormatOptions:
    """Options for an SDMX data/file format."""

    format: ClassVar[type[Format]]



import pytest

from sdmx.format.csv.common import Labels
from sdmx.format.csv.v1 import FormatOptions


class TestFormatOptions:
    def test_post_init(self) -> None:
        with pytest.raises(ValueError):
            FormatOptions(labels=Labels.name)



import pytest

from sdmx.model import common, v30
from sdmx.model.common import ConstraintRole, ConstraintRoleType
from sdmx.model.v30 import (
    Annotation,
    DataConstraint,
    GeoCodelistType,
    GeoFeatureSetCode,
    GeographicCodelist,
    GeoGridCode,
    GeoGridCodelist,
    HierarchicalCode,
    Hierarchy,
    HierarchyAssociation,
    Level,
    MetadataConstraint,
    MetadataProvider,
    MetadataProviderScheme,
)
from sdmx.testing import CompareTests

# §3.2: Basic structures


class TestAnnotation:
    def test_value(self) -> None:
        a0 = Annotation()

        # Value defaults None
        assert None is a0.value

        # Value can be set
        a0.value = "foo"

        # Value can be retrieved
        assert "foo" == a0.value

        # Value can be set through constructor
        a1 = Annotation(value="bar")
        assert "bar" == a1.value


# §4.3: Codelist


class TestGeoGridCode:
    def test_init(self):
        GeoGridCode(geo_cell="foo")


class TestGeoFeatureSetCode:
    def test_init(self):
        GeoFeatureSetCode(value="foo")


class TestGeographicCodelist:
    def test_init(self):
        cl = GeographicCodelist()

        assert GeoCodelistType.geographic == cl.geo_type


class TestGeoGridCodelist:
    def test_init(self):
        cl = GeoGridCodelist()

        assert GeoCodelistType.geogrid == cl.geo_type


# §4.7: OrganisationScheme


class TestMetadataProvider:
    def test_init(self):
        MetadataProvider()


class TestMetadataProviderScheme:
    def test_init(self):
        MetadataProviderScheme()


# §5.4: Data Set


class TestObservation(CompareTests):
    @pytest.fixture
    def obj(self) -> v30.Observation:
        return v30.Observation(
            attached_attribute={"FOO": common.AttributeValue(value="f1")},
            dimension=common.Key(BAR="b1"),
            group_keys={common.GroupKey(id="g1")},
            series_key=common.SeriesKey(),
            value=1.0,
            value_for=v30.Measure(id="m1"),
        )

    def test_compare(self, obj: v30.Observation, callback=None) -> None:
        """:py:`compare(…)` is :any:`False` when .value_for is changed.

        For other attributes, see test_common.TestBaseObservation.test_compare.
        """
        super().test_compare(
            obj, lambda obs: setattr(obs, "value_for", v30.Measure(id="m2"))
        )


# §8: Hierarchy


class TestLevel:
    def test_init(self):
        Level()


class TestHierarchicalCode:
    def test_init(self):
        HierarchicalCode()


class TestHierarchy:
    def test_init(self):
        Hierarchy()


class TestHierarchyAssociation:
    def test_init(self):
        HierarchyAssociation()


# §12.3: Constraints


_ROLE_PARAMS = [
    ConstraintRole(role=ConstraintRoleType.actual),
    ConstraintRole(role=ConstraintRoleType.allowable),
    ConstraintRoleType.actual,
    ConstraintRoleType.allowable,
    "actual",
    "allowable",
    pytest.param("foo", marks=pytest.mark.xfail(raises=KeyError)),
]


class TestDataConstraint:
    @pytest.mark.parametrize("role", _ROLE_PARAMS)
    def test_init(self, role):
        DataConstraint(role=role)


class TestMetadataConstraint:
    @pytest.mark.parametrize("role", _ROLE_PARAMS)
    def test_init(self, role):
        MetadataConstraint(role=role)



from datetime import datetime
from typing import TYPE_CHECKING

import sdmx
from sdmx.message import DataMessage, Header
from sdmx.model import v21 as model
from sdmx.model.v21 import AttributeValue, DataAttribute, DataSet, Key, Observation
from sdmx.testing import assert_pd_equal

if TYPE_CHECKING:
    from sdmx.testing.data import SpecimenCollection


def test_flat(specimen: "SpecimenCollection") -> None:
    # Create a bare Message
    msg = DataMessage()

    # Recreate the content from exr-flat.json
    header = Header(
        id="62b5f19d-f1c9-495d-8446-a3661ed24753",
        prepared=datetime.fromisoformat("2012-11-29T08:40:26+00:00"),
        sender=model.Agency(id="ECB"),
    )
    msg.header = header

    ds = DataSet()

    # Create a Key and attributes
    key = Key(
        FREQ="D",
        CURRENCY_DENOM="EUR",
        EXR_TYPE="SP00",
        EXR_SUFFIX="A",
        CURRENCY="NZD",
        TIME_PERIOD="2013-01-18",
    )
    assert msg.dataflow is not None
    [msg.dataflow.structure.dimensions.getdefault(id=d) for d in key.values]
    obs_status = DataAttribute(id="OBS_STATUS")
    attr = {"OBS_STATUS": AttributeValue(value_for=obs_status, value="A")}

    ds.obs.append(Observation(dimension=key, value=1.5931, attached_attribute=attr))

    key = key.copy(TIME_PERIOD="2013-01-21")
    ds.obs.append(Observation(dimension=key, value=1.5925, attached_attribute=attr))

    key = key.copy(CURRENCY="RUB", TIME_PERIOD="2013-01-18")
    ds.obs.append(Observation(dimension=key, value=40.3426, attached_attribute=attr))

    key = key.copy(TIME_PERIOD="2013-01-21")
    ds.obs.append(Observation(dimension=key, value=40.3000, attached_attribute=attr))

    msg.data.append(ds)

    # Write to pd.Dataframe
    df1 = sdmx.to_pandas(msg)

    with specimen("flat.json") as f:
        ref = sdmx.read_sdmx(f)
    df2 = sdmx.to_pandas(ref)

    assert_pd_equal(df1, df2)


def test_bare_series(specimen):
    with specimen("ng-ts.xml") as f:
        sdmx.read_sdmx(f)

    # TODO generate the series and observations



"""SDMX-ML 3.0.0 reader."""

import logging
from typing import Any

import sdmx.urn
from sdmx.format import Version
from sdmx.model import common, v30

from . import v21
from .common import BaseReference, NotReference, XMLEventReader

log = logging.getLogger(__name__)


class Reference(BaseReference):
    """Parse SDMX-ML 3.0 references."""

    @classmethod
    def info_from_element(cls, elem):
        try:
            result = sdmx.urn.match(elem.text)
            # If the URN doesn't specify an item ID, it is probably a reference to a
            # MaintainableArtefact, so target_id and id are the same
            result.update(target_id=result["item_id"] or result["id"])
        except (KeyError, ValueError):
            # Bare string that is the ID of e.g. a component (dimension)
            if id := (elem.text or "").strip():
                result = {"id": id, "target_id": id, "class": None, "package": None}
            else:
                raise NotReference()

        return result


class Reader(XMLEventReader):
    """SDMX-ML 3.0 reader."""

    xml_version = Version["3.0"]
    Reference = Reference


# Rewrite the v21.Reader collection of parsers to refer to SDMX-ML 3.0.0 namespaces
# instead of SDMX-ML 2.1
new_parsers = dict()
for (tag, event), func in v21.Reader.parser.items():
    # Construct a new tag using the same prefix (e.g. "str") and local name
    new_tag = Reader.format.qname(
        v21.Reader.format.ns_prefix(tag.namespace), tag.localname
    )
    # Store a reference to the same function
    new_parsers[(new_tag, event)] = func
# Replace the parser collection
Reader.parser = new_parsers

# Shorthand
start = Reader.start
end = Reader.end

# In SDMX-ML 3.0, individual classes of ItemScheme are collected in separate XML
# container elements. Skip all of these.
start(
    """
    str:AgencySchemes str:ConceptSchemes str:CustomTypeSchemes str:DataConstraints
    str:GeographicCodelists str:GeoGridCodelists str:Hierarchies
    str:NamePersonalisationSchemes str:RulesetSchemes str:TransformationSchemes
    str:UserDefinedOperatorSchemes str:ValueLists str:VtlMappingSchemes
    """
)(None)

# New qnames in SDMX-ML 3.0 parsed using existing methods from .reader.xml.v21
end("com:AnnotationValue str:GeoCell str:GridDefinition str:Value")(v21._text)
end("str:GeographicCodelist str:ValueList")(v21._itemscheme)
start("str:GeoFeatureSetCode str:GeoGridCode str:ValueItem", only=False)(
    v21._item_start
)
end("str:GeoFeatureSetCode str:GeoGridCode str:ValueItem", only=False)(v21._item_end)
start("str:Measure str:MetadataAttribute", only=False)(v21._component_start)
end("str:Measure str:MetadataAttribute", only=False)(v21._component_end)
end("str:MetadataAttributeList")(v21._cl)
end("str:DataConstraint")(v21._cc)
end("str:KeyValue")(v21._ms)
end("str:Observation")(v21._ar_kind)


@end("com:Link")
def _link(reader, elem) -> None:
    # TODO Once https://github.com/khaeru/sdmx/issues/228 is resolved, store on
    # IdentifiableArtefact or similar
    log.debug(f"Ignored: <com:Link> with attributes {elem.attrib}")


@end("str:Codelist")
def _cl(reader, elem):
    # Handle <str:Codelist>urn:…</str:Codelist>; occurs within <str:CodelistExtension>
    if len(elem) == 0 and elem.text:
        try:
            # Validate the contained URN, return to str
            urn = str(sdmx.urn.URN(elem.text))
        except Exception:
            pass  # Malformed, or not a URN
        else:
            reader.push(elem, urn)
            return

    # Use the .v21 reader function
    result = v21._itemscheme(reader, elem)
    # Attach CodelistExtensions
    result.extends = reader.pop_all(v30.CodelistExtension)
    return result


@end("str:CodelistExtension")
def _cl_ext(reader, elem):
    cs = reader.pop_all(v30.CodeSelection, subclass=True) or [None]
    assert 1 == len(cs)
    return v30.CodelistExtension(
        extends=reader.pop_resolved_ref("Codelist"),
        prefix=elem.attrib.get("prefix", None),
        selection=cs[0],
    )


@end("str:ExclusiveCodeSelection str:InclusiveCodeSelection")
def _code_selection(reader, elem):
    return reader.class_for_tag(elem.tag)(mv=reader.pop_all(reader.model.MemberValue))


@end("str:MemberValue")
def _mv(reader, elem):
    return reader.model.MemberValue(value=elem.text)


@end("str:GeoGridCodelist")
def _ggcl(reader, elem):
    result = v21._itemscheme(reader, elem)
    result.grid_definition = reader.pop_single("GridDefinition")
    return result


@end("str:GeoGridCode", only=False)
def _ggc_end(reader, elem):
    result = v21._item_end(reader, elem)
    result.geo_cell = reader.pop_single("GeoCell")
    return result


# §5.3: Data Structure Definition


@end("str:AttributeRelationship")
def _ar(reader: Reader, elem):
    dsd = reader.peek("current DSD")

    refs = reader.pop_all(reader.Reference)
    if not len(refs):
        return

    # Iterate over parsed references to Components
    args: dict[str, Any] = dict(dimensions=list())
    for ref in refs:
        # Use the <Ref id="..."> to retrieve a Component from the DSD
        if issubclass(ref.target_cls, common.DimensionComponent):
            component = dsd.dimensions.get(ref.target_id)
            args["dimensions"].append(component)
        elif ref.target_cls is common.Measure:
            # Since <str:AttributeList> occurs before <str:MeasureList>, this is
            # usually a forward reference. We *could* eventually resolve it to confirm
            # consistency (the referenced ID is same as the PrimaryMeasure.id), but
            # that doesn't affect the returned value, since PrimaryMeasureRelationship
            # has no attributes.
            return v30.ObservationRelationship()
        elif ref.target_cls is common.GroupDimensionDescriptor:
            args["group_key"] = dsd.group_dimensions[ref.target_id]

    ref = reader.pop_single("AttachmentGroup")
    if ref:
        args["group_key"] = dsd.group_dimensions[ref.target_id]

    if len(args["dimensions"]):
        return common.DimensionRelationship(**args)


# §5.4: Data Set


@end(":Value")
def _complex_value(reader: Reader, elem):
    try:
        reader.push("ComplexValue", common.InternationalString(reader.pop_all("Text")))
    except Exception:  # pragma: no cover
        raise NotImplementedError


@end(":Comp")
def _complex(reader: Reader, elem):
    ds = reader.get_single("DataSet")

    assert ds is not None
    da = ds.structured_by.attributes.getdefault(elem.attrib["id"])

    reader.stack.setdefault("Attributes", {-1: {}})

    reader.stack["Attributes"][-1][da.id] = common.AttributeValue(
        # See comment on AttributeValue.value
        value=reader.pop_all("ComplexValue"),  # type: ignore [arg-type]
        value_for=da,
    )


# §8: Hierarchy


@end("str:Hierarchy")
def _h(reader: Reader, elem):
    cls = reader.class_for_tag(elem.tag)
    return reader.maintainable(
        cls,
        elem,
        has_formal_levels=eval(elem.attrib["hasFormalLevels"].title()),
        codes={c.id: c for c in reader.pop_all(common.HierarchicalCode)},
        level=reader.pop_single(common.Level),
    )
